from pathlib import Path

from fpdf import FPDF
from fpdf.transitions import (
    SplitTransition,
    BlindsTransition,
    BoxTransition,
    WipeTransition,
    DissolveTransition,
    GlitterTransition,
    FlyTransition,
    PushTransition,
    CoverTransition,
    UncoverTransition,
    FadeTransition,
    Transition,
)
from test.conftest import assert_pdf_equal
import pytest

HERE = Path(__file__).resolve().parent


def test_transitions(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=120)
    pdf.add_page()
    pdf.text(x=40, y=150, text="Page 0")
    pdf.add_page(transition=SplitTransition("V", "O"))
    pdf.text(x=40, y=150, text="Page 1")
    pdf.add_page(transition=BlindsTransition("H"))
    pdf.text(x=40, y=150, text="Page 2")
    pdf.add_page(transition=BoxTransition("I"))
    pdf.text(x=40, y=150, text="Page 3")
    pdf.add_page(transition=WipeTransition(90))
    pdf.text(x=40, y=150, text="Page 4")
    pdf.add_page(transition=DissolveTransition())
    pdf.text(x=40, y=150, text="Page 5")
    pdf.add_page(transition=GlitterTransition(315))
    pdf.text(x=40, y=150, text="Page 6")
    pdf.add_page(transition=FlyTransition("H"))
    pdf.text(x=40, y=150, text="Page 7")
    pdf.add_page(transition=PushTransition(270))
    pdf.text(x=40, y=150, text="Page 8")
    pdf.add_page(transition=CoverTransition(270))
    pdf.text(x=40, y=150, text="Page 9")
    pdf.add_page(transition=UncoverTransition(270))
    pdf.text(x=40, y=150, text="Page 10")
    pdf.add_page(transition=FadeTransition())
    pdf.text(x=40, y=150, text="Page 11")
    assert_pdf_equal(pdf, HERE / "transitions.pdf", tmp_path)


def test_transition_errors():
    pdf = FPDF()
    pdf.set_font("Helvetica", size=120)
    with pytest.raises(NotImplementedError):
        Transition().serialize()

    with pytest.raises(ValueError):
        pdf.add_page(transition=SplitTransition("A", "B"))

    with pytest.raises(ValueError):
        pdf.add_page(transition=SplitTransition("H", "B"))

    with pytest.raises(ValueError):
        pdf.add_page(transition=BlindsTransition("A"))

    with pytest.raises(ValueError):
        pdf.add_page(transition=BoxTransition("A"))

    with pytest.raises(ValueError):
        pdf.add_page(transition=WipeTransition(-1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=GlitterTransition(-1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=FlyTransition("A", -1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=FlyTransition("V", -1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=PushTransition(-1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=CoverTransition(-1))

    with pytest.raises(ValueError):
        pdf.add_page(transition=UncoverTransition(-1))



from pathlib import Path

from fpdf import FPDF
from fpdf.enums import OutputIntentSubType
from fpdf.output import PDFICCProfile, OutputIntentDictionary

import pytest
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_output_intents_coerce():
    """
    Make sure that the coerce function returns expected subtype,
    or raises ValueError on not expected input
    """
    assert OutputIntentSubType.coerce("PDFA") == OutputIntentSubType.PDFA
    assert OutputIntentSubType.coerce("pdfa") == OutputIntentSubType.PDFA
    assert OutputIntentSubType.coerce("PDFX") == OutputIntentSubType.PDFX
    assert OutputIntentSubType.coerce("ISOPDF") == OutputIntentSubType.ISOPDF
    with pytest.raises(ValueError):
        assert OutputIntentSubType.coerce("BXXX")


def test_output_intents_properties():
    """
    Make sure the properties in PDF return the correct
    value as of the originating Output Intent.
    """
    pdf = FPDF()

    assert not pdf.output_intents

    pdf.add_output_intent(OutputIntentSubType.PDFA, "sRGB")  # should add PDFA
    pdf.add_output_intent(OutputIntentSubType.PDFX, "AdobeRGB")  # should add PDFX
    pdf.add_output_intent(OutputIntentSubType.ISOPDF, "AdobeRGB")  # should add ISOPDF

    assert (
        sum(
            1 for item in pdf.output_intents if isinstance(item, OutputIntentDictionary)
        )
        == 3
    ), "Array does not have exactly 3 objects of type OutputIntentDictionary"


def test_output_intents(tmp_path):
    """
    Make sure the Output Intents is set in PDF.
    """
    doc = FPDF()
    with open(HERE / "sRGB2014.icc", "rb") as iccp_file:
        icc_profile = PDFICCProfile(
            contents=iccp_file.read(), n=3, alternate="DeviceRGB"
        )

    doc.add_output_intent(
        OutputIntentSubType.PDFA,
        "sRGB",
        "IEC 61966-2-1:1999",
        "http://www.color.org",
        icc_profile,
        "sRGB2014 (v2)",
    )
    doc.set_lang("de")
    doc.add_page()
    doc.set_font("Helvetica", size=20)
    for i, color in enumerate(
        (
            (255, 100, 100),
            (255, 255, 100),
            (255, 100, 255),
            (250, 250, 250),
            (0, 0, 0),
        )
    ):
        doc.set_text_color(*color)
        doc.text(20, 20 + 10 * i, f"{color}")
    assert_pdf_equal(doc, HERE / "text_color_with_one_output_intent.pdf", tmp_path)


def test_output_intents_without_optionals(tmp_path):
    """
    Make sure the Output Intent is set in PDF.
    """
    doc = FPDF()
    doc.add_output_intent(OutputIntentSubType.PDFA, "somethingStrange")
    doc.set_lang("de")
    doc.add_page()
    doc.set_font("Helvetica", size=20)
    for i, color in enumerate(
        (
            (255, 100, 100),
            (255, 255, 100),
            (255, 100, 255),
            (250, 250, 250),
            (0, 0, 0),
        )
    ):
        doc.set_text_color(*color)
        doc.text(20, 20 + 10 * i, f"{color}")
    assert_pdf_equal(
        doc, HERE / "text_color_with_one_output_intent_without_optionals.pdf", tmp_path
    )


def test_two_output_intents(tmp_path):
    """
    Make sure the Output Intents is set in PDF.
    """
    doc = FPDF()
    with open(HERE / "sRGB2014.icc", "rb") as iccp_file:
        icc_profile = PDFICCProfile(
            contents=iccp_file.read(), n=3, alternate="DeviceRGB"
        )
    doc.add_output_intent(
        OutputIntentSubType.PDFA,
        "sRGB",
        "IEC 61966-2-1:1999",
        "http://www.color.org",
        icc_profile,
        "sRGB2014 (v2)",
    )
    doc.add_output_intent(
        OutputIntentSubType.ISOPDF,
        "somethingStrange",
    )
    doc.set_lang("de")
    doc.add_page()
    doc.set_font("Helvetica", size=20)
    for i, color in enumerate(
        (
            (255, 100, 100),
            (255, 255, 100),
            (255, 100, 255),
            (250, 250, 250),
            (0, 0, 0),
        )
    ):
        doc.set_text_color(*color)
        doc.text(20, 20 + 10 * i, f"{color}")
    assert_pdf_equal(doc, HERE / "text_color_with_two_output_intents.pdf", tmp_path)


def test_two_equal_output_intents_raises():
    """
    Make sure the second Output Intent raises ValueError.
    """
    doc = FPDF()
    with open(HERE / "sRGB2014.icc", "rb") as iccp_file:
        icc_profile = PDFICCProfile(
            contents=iccp_file.read(), n=3, alternate="DeviceRGB"
        )
    doc.add_output_intent(
        OutputIntentSubType.PDFA,
        "sRGB",
        "IEC 61966-2-1:1999",
        "http://www.color.org",
        icc_profile,
        "sRGB2014 (v2)",
    )
    with pytest.raises(ValueError):
        assert doc.add_output_intent(OutputIntentSubType.PDFA, "somethingStrange")


def test_without_output_intents(tmp_path):
    """
    Make sure the Output Intents is not set in PDF.
    """
    doc = FPDF()
    doc.set_lang("de")
    doc.add_page()
    doc.set_font("Helvetica", size=20)
    for i, color in enumerate(
        (
            (255, 100, 100),
            (255, 255, 100),
            (255, 100, 255),
            (250, 250, 250),
            (0, 0, 0),
        )
    ):
        doc.set_text_color(*color)
        doc.text(20, 20 + 10 * i, f"{color}")
    assert_pdf_equal(doc, HERE / "text_color_without_output_intent.pdf", tmp_path)



import pytest
import sys

from pathlib import Path
from fpdf import FPDF
from test.conftest import assert_pdf_equal, EPOCH

HERE = Path(__file__).resolve().parent


def _make_text_file(tmp_path: Path) -> Path:
    return _make_text_file_with_name(tmp_path, "helloworld.txt")


def _make_text_file_with_name(tmp_path: Path, filename: str) -> Path:
    p = tmp_path / filename
    p.write_text("Hello world", encoding="utf-8")
    return p


def test_embed_file_self(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(_make_text_file(tmp_path), modification_date=False)
    assert_pdf_equal(pdf, HERE / "embed_file_self.pdf", tmp_path)


def test_embed_file_parens(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(
        _make_text_file_with_name(tmp_path, "helloworld(1).txt"),
        modification_date=False,
    )
    assert_pdf_equal(pdf, HERE / "embed_file_parens.pdf", tmp_path)


def test_embed_file_single_parens(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(
        _make_text_file_with_name(tmp_path, "helloworld(1.txt"),
        modification_date=False,
        desc="with (parens",
    )
    assert_pdf_equal(pdf, HERE / "embed_file_single_parens.pdf", tmp_path)


@pytest.mark.skipif(
    sys.platform in ("cygwin", "win32") and sys.version_info[:2] == (3, 14),
    reason="Skipped on Windows with Python 3.14 due to zlib compressed data differences",
)
def test_embed_file_all_optionals(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(
        str(_make_text_file(tmp_path)),
        desc="Source Python code",
        creation_date=EPOCH,
        modification_date=EPOCH,
        compress=True,
        checksum=True,
    )
    assert_pdf_equal(pdf, HERE / "embed_file_all_optionals.pdf", tmp_path)


def test_embed_file_from_bytes(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(bytes=b"Hello world!", basename="hello_world.txt")
    assert_pdf_equal(pdf, HERE / "embed_file_from_bytes.pdf", tmp_path)


def test_file_attachment_annotation(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.file_attachment_annotation(
        _make_text_file(tmp_path), modification_date=False, x=50, y=50
    )
    assert_pdf_equal(pdf, HERE / "file_attachment_annotation.pdf", tmp_path)


def test_embed_file_invalid_params():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(ValueError):
        pdf.embed_file(__file__, bytes=b"...")
    with pytest.raises(ValueError):
        pdf.embed_file(__file__, basename="file.txt")
    with pytest.raises(ValueError):
        pdf.embed_file(bytes=b"...")
    with pytest.raises(ValueError):
        pdf.embed_file(basename="file.txt")


def test_embed_file_duplicate(tmp_path):
    EMBEDDED_FILE = _make_text_file(tmp_path)
    pdf = FPDF()
    pdf.add_page()
    pdf.embed_file(EMBEDDED_FILE)
    with pytest.raises(ValueError):
        pdf.embed_file(EMBEDDED_FILE)



import io
import plotly.graph_objects as go
import numpy as np
from fpdf import FPDF

np.random.seed(1)

N = 100
x = np.random.rand(N)
y = np.random.rand(N)
colors = np.random.rand(N)
sz = np.random.rand(N) * 30

fig = go.Figure()
fig.add_trace(
    go.Scatter(
        x=x,
        y=y,
        mode="markers",
        marker=go.scatter.Marker(
            size=sz, color=colors, opacity=0.6, colorscale="Viridis"
        ),
    )
)

# Convert the figure to png using kaleido
image_data = fig.to_image(format="png", engine="kaleido")

# Create an io.BytesIO object which can be used by FPDF2
image = io.BytesIO(image_data)
pdf = FPDF()
pdf.add_page()
pdf.image(image, w=pdf.epw)  # Width of the image is equal to the width of the page
pdf.output("plotly_demo.pdf")



from fpdf import FPDF
import plotly.express as px

fig = px.bar(x=["a", "b", "c"], y=[1, 3, 2])
fig.write_image("figure.svg")

pdf = FPDF()
pdf.add_page()
pdf.image("figure.svg", w=pdf.epw)
pdf.output("plotly.pdf")



#!/usr/bin/env python3
# USAGE: ./csv2table.py color_srgb.csv
import csv, sys
from fpdf import FPDF, FontFace
from fpdf.drawing_primitives import color_from_hex_string

pdf = FPDF()
pdf.add_page()
pdf.set_font("Times", size=22)
with pdf.table() as table:
    with open(sys.argv[1], encoding="utf-8") as csv_file:
        reader = csv.reader(csv_file, delimiter=",")
        for i, row in enumerate(reader):
            style = None
            if i > 0:
                # We color the row based on the hexadecimal code in the 2nd column:
                style = FontFace(fill_color=color_from_hex_string(row[1]))
            table.row(row, style=style)
pdf.output("from-csv.pdf")



import binascii
from glob import glob
from pathlib import Path

import pytest

import fpdf

from test.conftest import assert_pdf_equal, ensure_rss_memory_below, time_execution

HERE = Path(__file__).resolve().parent


def test_load_text_file():
    file = HERE / "__init__.py"
    contents = '"""This package contains image tests"""\n'
    bc = contents.encode()

    resource = fpdf.image_parsing.load_image(str(file)).getvalue()
    # loaded a text file in binary mode, may contain DOS style line endings.
    resource = resource.replace(b"\r\n", b"\n")
    assert bytes(resource) == bc


def test_load_base64_data(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(
        "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQBAMAAADt3eJSAAAAMFBMVEU0OkArMjhobHEoPUPFEBIuO0L+AAC2FBZ2JyuNICOfGx7xAwT"
        "jCAlCNTvVDA1aLzQ3COjMAAAAVUlEQVQI12NgwAaCDSA0888GCItjn0szWGBJTVoGSCjWs8TleQCQYV95evdxkFT8Kpe0PLDi5WfKd4LUsN5zS1sKFolt8bwAZrCa"
        "GqNYJAgFDEpQAAAzmxafI4vZWwAAAABJRU5ErkJggg=="
    )
    assert_pdf_equal(pdf, HERE / "load_base64_data.pdf", tmp_path)


def test_load_invalid_base64_data():
    pdf = fpdf.FPDF()
    pdf.add_page()
    with pytest.raises(binascii.Error):
        pdf.image("data:image/png;base64,GARBAGE")


# ensure memory usage does not get too high - this value depends on Python version:
@ensure_rss_memory_below(mib=6)
def test_share_images_cache(tmp_path):
    image_cache = None

    def build_pdf_with_big_images():
        nonlocal image_cache
        pdf = fpdf.FPDF()
        if image_cache is None:
            image_cache = pdf.image_cache
        else:
            pdf.image_cache = image_cache
        pdf.add_page()
        for img_path in glob(f"{HERE}/png_images/*.png"):
            pdf.image(img_path, h=pdf.eph)
        with (tmp_path / "out.pdf").open("wb") as pdf_file:
            pdf.output(pdf_file)
        # Reset the "usages" count:
        image_cache.reset_usages()

    with time_execution() as duration:
        build_pdf_with_big_images()
    first_time_duration = duration.seconds

    with time_execution() as duration:
        build_pdf_with_big_images()
    assert duration.seconds < first_time_duration / 2



import itertools
from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.errors import FPDFException
from fpdf.fonts import CORE_FONTS, CORE_FONTS_CHARWIDTHS
from test.conftest import assert_pdf_equal, assert_same_file

HERE = Path(__file__).resolve().parent


def test_no_set_font():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        pdf.text(10, 10, "Hello World!")
    expected_msg = "No font set, you need to call set_font() beforehand"
    assert str(error.value) == expected_msg


def test_set_unknown_font():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as e:
        pdf.set_font("Dummy")
    assert (
        str(e.value)
        == "Undefined font: dummy - Use built-in fonts or FPDF.add_font() beforehand"
    )


def test_set_unknown_style():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(ValueError) as e:
        pdf.set_font("Times", style="bold")
    assert (
        str(e.value)
        == "Unknown style provided (only B/I/S/U letters are allowed): BDLO"
    )


def test_set_builtin_font(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    builtin_fonts = sorted(f for f in CORE_FONTS if not f.endswith(("B", "I", "BI")))
    for i, font_name in enumerate(builtin_fonts):
        styles = (
            ("",) if font_name in ("symbol", "zapfdingbats") else ("", "B", "I", "BI")
        )
        for j, style in enumerate(styles):
            pdf.set_font(font_name.capitalize(), style, size=36)
            pdf.set_font(font_name.lower(), style, size=36)
            pdf.text(0, 10 + 40 * i + 10 * j, "Hello World!")
    assert_pdf_equal(pdf, HERE / "fonts_set_builtin_font.pdf", tmp_path)


def test_issue_66(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", style="B", size=14)
    pdf.cell(text="ABC")
    pdf.set_font("Times", size=10)
    pdf.cell(text="DEF")
    # Setting the font to an already used one used to remove the text!
    pdf.set_font("Times", style="B", size=14)
    assert_pdf_equal(pdf, HERE / "fonts_issue_66.pdf", tmp_path)


def test_set_font_aliases_as_font():
    """Test if font aliases are being converted to their alternatives."""
    pdf = FPDF()
    pdf.add_page()
    aliases = ["ARIAL", "Arial", "arial", "couriernew", "timesnewroman"]
    alternatives = ["helvetica", "helvetica", "helvetica", "courier", "times"]
    for alias, alternative in zip(aliases, alternatives):
        # Test if warning get's emitted
        with pytest.warns(
            DeprecationWarning,
            match=f"Substituting font {alias.lower()} by core font {alternative} - This is deprecated since v2.7.8, and will soon be removed",
        ) as record:
            pdf.set_font(alias)

        assert len(record) == 1
        assert_same_file(record[0].filename, __file__)

        # Test if font family is set correctly
        assert pdf.font_family == alternative
    # Test if the fonts were added in this order and without duplicates:
    # helvetica, courier, times
    assert [*pdf.fonts] == ["helvetica", "courier", "times"]


def test_set_font_core_font_attributes():
    """Test if the attributes of added core fonts are correct"""
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("courier")
    pdf.set_font("times")

    # Test for the font attributes
    assert pdf.fonts["courier"].i == 1
    assert pdf.fonts["courier"].type == "core"
    assert pdf.fonts["courier"].name == "Courier"
    assert pdf.fonts["courier"].up == -100
    assert pdf.fonts["courier"].ut == 50
    assert pdf.fonts["courier"].cw == CORE_FONTS_CHARWIDTHS["courier"]
    assert pdf.fonts["courier"].fontkey == "courier"
    assert pdf.fonts["courier"].emphasis == 0

    assert pdf.fonts["times"].i == 2
    assert pdf.fonts["times"].type == "core"
    assert pdf.fonts["times"].name == "Times-Roman"
    assert pdf.fonts["times"].up == -100
    assert pdf.fonts["times"].ut == 50
    assert pdf.fonts["times"].cw == CORE_FONTS_CHARWIDTHS["times"]
    assert pdf.fonts["times"].fontkey == "times"
    assert pdf.fonts["times"].emphasis == 0


def test_set_font_styles():
    """Test the different font styles "B", "I" and "U" and combinations."""
    pdf = FPDF()
    pdf.add_page()

    # Generate all possible combinations of "B", "I" and "U" -> "B", "BI", "BUI" ...
    # including "" (no style)
    styles = [
        "".join(style) for i in range(4) for style in itertools.permutations("BUI", i)
    ]

    for style in styles:
        pdf.set_font("Times", style=style)

        # Test if underline is set correctly
        assert pdf.underline == int("U" in style)

        # pylint: disable=redefined-loop-name
        # Test if style is set correctly
        style = style.replace("U", "")
        if style == "IB":
            style = "BI"
        assert pdf.font_style == style


def test_set_font_zapfdingbats_symbol_with_style():
    """Test the fonts zapfdingbats and symbol with a style. This should emit a
    warning, as these fonts don't have a style."""
    pdf = FPDF()
    pdf.add_page()

    # Generate all possible combinations of "B", "I" and "U" -> "B", "BI", "BUI" ...
    # excluding "" (no style)
    styles = [
        "".join(style)
        for i in range(1, 4)
        for style in itertools.permutations("BUI", i)
    ]
    for family in ("zapfdingbats", "symbol"):
        for style in styles:
            if "B" in style or "I" in style:
                with pytest.warns(
                    UserWarning,
                    match=f"Built-in font {family} only has a single 'style' and "
                    f"can't be bold or italic",
                ) as record:
                    pdf.set_font(family, style=style)

                    # Test if style is set correctly (== no style)
                    assert pdf.font_style == ""

                assert len(record) == 1
                assert_same_file(record[0].filename, __file__)

                # Test if underline is set correctly
                assert pdf.underline == int("U" in style)



# pylint: disable=protected-access
from os import devnull
from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.actions import GoToAction
from fpdf.encryption import StandardSecurityHandler as sh
from fpdf.enums import AccessPermission, EncryptionMethod
from fpdf.errors import FPDFException
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent

XMP_METADATA = """<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="fpdf2">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description rdf:about="">
      <dc:title xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Alt>
          <rdf:li xml:lang="x-default">My document title</rdf:li>
        </rdf:Alt>
      </dc:title>
    </rdf:Description>
    <rdf:Description rdf:about="">
      <dc:description xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Alt>
          <rdf:li xml:lang="x-default">This is a test document for fpdf2 with XMP metadata</rdf:li>
        </rdf:Alt>
      </dc:description>
    </rdf:Description>
    <rdf:Description rdf:about="">
      <dc:creator xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Seq>
          <rdf:li>Lucas Cimon</rdf:li>
        </rdf:Seq>
      </dc:creator>
    </rdf:Description>
    <rdf:Description xmlns:pdf="http://ns.adobe.com/pdf/1.3/" rdf:about="" pdf:Keywords="test data pdf fpdf2"/>
    <rdf:Description xmlns:pdf="http://ns.adobe.com/pdf/1.3/" rdf:about="" pdf:Producer="py-pdf/fpdf2.X.Y"/>
    <rdf:Description xmlns:xmp="http://ns.adobe.com/xap/1.0/" rdf:about="" xmp:CreatorTool="fpdf2"/>
  </rdf:RDF>
</x:xmpmeta>"""


def test_encryption_rc4(tmp_path):
    pdf = FPDF()
    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(owner_password="fpdf2", permissions=AccessPermission.all())
    assert_pdf_equal(pdf, HERE / "encryption_rc4.pdf", tmp_path)


def test_encryption_rc4_permissions(tmp_path):
    pdf = FPDF()
    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        permissions=AccessPermission.PRINT_LOW_RES | AccessPermission.PRINT_HIGH_RES,
    )
    assert_pdf_equal(pdf, HERE / "encryption_rc4_permissions.pdf", tmp_path)


def test_no_encryption(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return pdf._default_file_id(bytearray([0xFF]))

    pdf.file_id = custom_file_id
    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        encryption_method=EncryptionMethod.NO_ENCRYPTION,
        permissions=AccessPermission.none(),
    )
    assert_pdf_equal(pdf, HERE / "no_encryption.pdf", tmp_path)


def test_encryption_empty_user_password(tmp_path):
    pdf = FPDF()
    pdf.set_encryption(owner_password="fpdf2", user_password="")
    assert_pdf_equal(pdf, HERE / "encryption_empty_user_password.pdf", tmp_path)


def test_encryption_rc4_user_password(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return pdf._default_file_id(bytearray([0xFF]))

    pdf.file_id = custom_file_id
    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        user_password="654321",
        permissions=AccessPermission.PRINT_LOW_RES | AccessPermission.PRINT_HIGH_RES,
    )
    assert_pdf_equal(pdf, HERE / "encryption_rc4_user_password.pdf", tmp_path)


def test_encryption_aes128(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return pdf._default_file_id(bytearray([0xFF]))

    pdf.file_id = custom_file_id

    def fixed_iv(size):
        return bytearray(size)

    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        encryption_method=EncryptionMethod.AES_128,
        permissions=AccessPermission.none(),
    )
    pdf._security_handler.get_random_bytes = fixed_iv
    assert_pdf_equal(pdf, HERE / "encryption_aes128.pdf", tmp_path)


def test_encrypt_metadata(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return "<AC2718D5DA802D34E7F97EEF0A0B52C5><AC2718D5DA802D34E7F97EEF0A0B52C5>"

    pdf.file_id = custom_file_id
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        encrypt_metadata=True,
    )
    pdf.set_xmp_metadata(XMP_METADATA)
    assert_pdf_equal(pdf, HERE / "encrypt_metadata.pdf", tmp_path)


def test_encrypt_font(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(
        "Quicksand", style="", fname=HERE.parent / "fonts" / "Quicksand-Regular.otf"
    )
    pdf.add_font(
        "Quicksand", style="B", fname=HERE.parent / "fonts" / "Quicksand-Bold.otf"
    )
    pdf.add_font(
        "Quicksand", style="I", fname=HERE.parent / "fonts" / "Quicksand-Italic.otf"
    )
    pdf.set_font("Quicksand", size=32)
    text = (
        "Lorem ipsum dolor, **consectetur adipiscing** elit,"
        " eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True)
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    pdf.set_encryption(owner_password="fpdf2")
    assert_pdf_equal(pdf, HERE / "encrypt_fonts.pdf", tmp_path)


def test_encryption_with_hyperlink(tmp_path):  # issue 672
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica")
    pdf.cell(text="hyperlink", link="https://github.com/py-pdf/fpdf2")
    pdf.set_encryption(owner_password="fpdf2")
    assert_pdf_equal(pdf, HERE / "encryption_with_hyperlink.pdf", tmp_path)


def test_encrypt_outline(tmp_path):  # issue 732
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica")
    pdf.start_section("Title")
    pdf.start_section("Subtitle", level=1)
    pdf.set_encryption(owner_password="fpdf2")
    assert_pdf_equal(pdf, HERE / "encrypt_outline.pdf", tmp_path)


def test_encryption_aes256(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return pdf._default_file_id(bytearray([0xFF]))

    pdf.file_id = custom_file_id

    def fixed_iv(size):
        return bytearray(size)

    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.text(50, 50, "Some text")
    pdf.ink_annotation(
        [(40, 50), (70, 25), (100, 50), (70, 75), (40, 50)],
        title="Lucas",
        text="Some encrypted annotation",
    )
    pdf.set_encryption(
        owner_password="fpdf2",
        encryption_method=EncryptionMethod.AES_256,
        permissions=AccessPermission.none(),
    )
    pdf._security_handler.get_random_bytes = fixed_iv
    assert_pdf_equal(pdf, HERE / "encryption_aes256.pdf", tmp_path)


def test_encryption_aes256_with_user_password(tmp_path):
    pdf = FPDF()

    def custom_file_id():
        return pdf._default_file_id(bytearray([0xFF]))

    pdf.file_id = custom_file_id

    def fixed_iv(size):
        return bytearray(size)

    pdf.set_author("author")
    pdf.set_subject("string to be encrypted")
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.cell(text="hello world")
    pdf.set_encryption(
        owner_password="fpdf2",
        user_password="1" * 1000,
        encryption_method=EncryptionMethod.AES_256,
        permissions=AccessPermission.all(),
    )
    pdf._security_handler.get_random_bytes = fixed_iv
    assert_pdf_equal(pdf, HERE / "encryption_aes256_user_password.pdf", tmp_path)


def test_blank_owner_password():
    pdf = FPDF()
    pdf.set_encryption(
        owner_password="",
        encryption_method=EncryptionMethod.AES_256,
        permissions=AccessPermission.none(),
    )
    with pytest.raises(FPDFException) as e:
        pdf.output(devnull)
    assert str(e.value) == "Invalid owner password "


def test_password_prep():
    """
    The PDF standard requires the passwords to be prepared using the stringprep algorithm
    using the SASLprep as per RFC 4013
    https://datatracker.ietf.org/doc/html/rfc4013
    Those assertions are bases on the examples section of the RFC
    """
    assert sh.prepare_string("I\xadX") == b"IX"  # SOFT HYPHEN mapped to nothing
    assert sh.prepare_string("user") == b"user"  # no transformation
    assert sh.prepare_string("USER") == b"USER"  # case preserved
    assert sh.prepare_string("\xaa") == b"a"  # output is NFKC, input in ISO 8859-1
    assert sh.prepare_string("\u2168") == b"IX"  # output is NFKC, will match #1
    with pytest.raises(FPDFException) as e:
        sh.prepare_string("\x07")  # Error - prohibited character
    assert str(e.value) == "The password  contains prohibited characters"
    with pytest.raises(FPDFException) as e:
        sh.prepare_string("\u0627\x31")  # Error - bidirectional check
    assert sh.prepare_string("A" * 300) == b"A" * 127  # test cap 127 chars


def test_encryption_unicode(tmp_path):
    "Issue #933"
    pdf = FPDF()
    pdf.set_author("Thai")
    pdf.set_subject("ทดสอบภาษาไทย")
    pdf.add_page()
    pdf.set_text_shaping()
    pdf.add_font("Garuda", fname=HERE.parent / "fonts" / "Garuda.ttf")
    pdf.set_font("Garuda", size=12)
    pdf.start_section("ทดสอบภาษาไทย")
    pdf.cell(
        text="สวัสดี ทดสอบภาษาไทย กีกี้ กาก้า ก๋า อ้า อ้ำ ฤาษี ทุ่มทุน อุ้งอุ๋ง น้ำใจ ฯลฯ ญาญ่า ฐาน ฎีกา ฏฒัฯนณ ภัทร์ สิทธิ์"
    )
    pdf.set_encryption(owner_password="fpdf2")
    assert_pdf_equal(pdf, HERE / "encryption_unicode.pdf", tmp_path)


def test_goto_named_destination(tmp_path):
    pdf = FPDF()
    pdf.set_encryption(owner_password="fpdf2")
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    x, y, text = 80, 140, "GoTo action"
    pdf.text(x=x, y=y, text=text)
    pdf.add_action(
        GoToAction(dest="#page2"),
        x=x,
        y=y - pdf.font_size,
        w=pdf.get_string_width(text),
        h=pdf.font_size,
    )
    pdf.add_page()
    pdf.add_link(y=140, name="page2")
    pdf.text(x=80, y=140, text="Page 2")
    assert_pdf_equal(
        pdf,
        HERE / "encryption_goto_named_destination.pdf",
        tmp_path,
    )



from pathlib import Path

from fpdf import Align, FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
IMAGE_PATH = HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png"
SVG_PATH = HERE / "../svg/svg_sources/SVG_logo_fixed_dimensions.svg"


def test_image_x_align_center(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.image(IMAGE_PATH, x="C")
    pdf.image(IMAGE_PATH, x=Align.C)
    pdf.add_page()
    pdf.image(SVG_PATH, x="C")
    pdf.image(SVG_PATH, x=Align.C)
    assert_pdf_equal(pdf, HERE / "image_x_align_center.pdf", tmp_path)


def test_image_x_align_right(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.image(IMAGE_PATH, x="R")
    pdf.image(IMAGE_PATH, x=Align.R)
    pdf.add_page()
    pdf.image(SVG_PATH, x="R")
    pdf.image(SVG_PATH, x=Align.R)
    assert_pdf_equal(pdf, HERE / "image_x_align_right.pdf", tmp_path)



from pathlib import Path

import pytest

from fpdf import FPDF, drawing
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_graphics_context(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    pdf.set_text_color(0x00, 0xFF, 0x00)
    pdf.set_fill_color(0xFF, 0x88, 0xFF)
    pdf.set_y(20)
    pdf.cell(text="outer 01", new_x="LMARGIN", new_y="NEXT", fill=True)
    with pdf.local_context():
        pdf.set_font("courier", style="BIU", size=30)
        pdf.set_text_color(0xFF, 0x00, 0x00)
        pdf.set_fill_color(0xFF, 0xFF, 0x00)
        pdf.cell(text="inner 01", new_x="LMARGIN", new_y="NEXT", fill=True)
        pdf.set_x(70)
        with pdf.rotation(30, pdf.get_x(), pdf.get_y()):
            pdf.set_fill_color(0x00, 0xFF, 0x00)
            pdf.cell(text="inner 02", new_x="LMARGIN", new_y="NEXT", fill=True)
        pdf.set_stretching(150)
        pdf.cell(text="inner 03", new_x="LMARGIN", new_y="NEXT", fill=True)
    pdf.cell(text="outer 02", new_x="LMARGIN", new_y="NEXT", fill=True)
    assert_pdf_equal(pdf, HERE / "graphics_context.pdf", tmp_path)


def test_change_settings():
    """
    Make sure all the the set_xxx() methods stored in the graphics context
    do the right thing, both when changing something or using the same value
    again (the latter will often execute a separate, shortened code path).
    """
    red = (255, 0, 0)
    blue = (0, 255, 0)
    green = (0, 0, 255)

    # draw color
    pdf = FPDF()
    # verify default
    tgt_draw_color = pdf.DEFAULT_DRAW_COLOR
    assert pdf.draw_color == tgt_draw_color, (
        f"pdf.draw_color ({pdf.draw_color})" f" != tgt_draw_color ({tgt_draw_color})"
    )
    # change to red
    tgt_draw_color = drawing.DeviceRGB(*[c / 255 for c in red])
    pdf.set_draw_color(*red)
    assert pdf.draw_color == tgt_draw_color, (
        f"pdf.draw_color ({pdf.draw_color})" f" != tgt_draw_color ({tgt_draw_color})"
    )
    # stays the same
    pdf.set_draw_color(*red)
    assert pdf.draw_color == tgt_draw_color, (
        f"pdf.draw_color ({pdf.draw_color})" f" != tgt_draw_color ({tgt_draw_color})"
    )

    # fill color
    pdf = FPDF()
    # verify default
    tgt_fill_color = pdf.DEFAULT_FILL_COLOR
    assert pdf.fill_color == tgt_fill_color, (
        f"pdf.fill_color ({pdf.fill_color})" f" != tgt_fill_color ({tgt_fill_color})"
    )
    # change to green
    tgt_fill_color = drawing.DeviceRGB(*[c / 255 for c in green])
    pdf.set_fill_color(*green)
    assert pdf.fill_color == tgt_fill_color, (
        f"pdf.fill_color ({pdf.fill_color})" f" != tgt_fill_color ({tgt_fill_color})"
    )
    # stays the same
    pdf.set_fill_color(*green)
    assert pdf.fill_color == tgt_fill_color, (
        f"pdf.fill_color ({pdf.fill_color})" f" != tgt_fill_color ({tgt_fill_color})"
    )

    # text color
    pdf = FPDF()
    # verify default
    tgt_text_color = pdf.DEFAULT_TEXT_COLOR
    assert pdf.text_color == tgt_text_color, (
        f"pdf.text_color ({pdf.text_color})" f" != tgt_text_color ({tgt_text_color})"
    )
    # change to blue
    tgt_text_color = drawing.DeviceRGB(*[c / 255 for c in blue])
    pdf.set_text_color(*blue)
    assert pdf.text_color == tgt_text_color, (
        f"pdf.text_color ({pdf.text_color})" f" != tgt_text_color ({tgt_text_color})"
    )
    # stays the same
    pdf.set_text_color(*blue)
    assert pdf.text_color == tgt_text_color, (
        f"pdf.text_color ({pdf.text_color})" f" != tgt_text_color ({tgt_text_color})"
    )

    # underline
    # no setter method for this one

    # font_stretching
    pdf = FPDF()
    # verify default
    tgt_font_stretching = 100
    assert pdf.font_stretching == tgt_font_stretching, (
        f"pdf.font_stretching ({pdf.font_stretching})"
        f" != tgt_font_stretching ({tgt_font_stretching})"
    )
    # change
    tgt_font_stretching = 120
    pdf.set_stretching(tgt_font_stretching)
    assert pdf.font_stretching == tgt_font_stretching, (
        f"pdf.font_stretching ({pdf.font_stretching})"
        f" != tgt_font_stretching ({tgt_font_stretching})"
    )
    # stays the same
    pdf.set_stretching(tgt_font_stretching)
    assert pdf.font_stretching == tgt_font_stretching, (
        f"pdf.font_stretching ({pdf.font_stretching})"
        f" != tgt_font_stretching ({tgt_font_stretching})"
    )

    # char_spacing
    pdf = FPDF()
    # verify default
    tgt_char_spacing = 0
    assert pdf.char_spacing == tgt_char_spacing, (
        f"pdf.char_spacing ({pdf.char_spacing})"
        f" != tgt_char_spacing ({tgt_char_spacing})"
    )
    # change
    tgt_char_spacing = 5
    pdf.set_char_spacing(tgt_char_spacing)
    assert pdf.char_spacing == tgt_char_spacing, (
        f"pdf.char_spacing ({pdf.char_spacing})"
        f" != tgt_char_spacing ({tgt_char_spacing})"
    )
    # stays the same
    pdf.set_char_spacing(tgt_char_spacing)
    assert pdf.char_spacing == tgt_char_spacing, (
        f"pdf.char_spacing ({pdf.char_spacing})"
        f" != tgt_char_spacing ({tgt_char_spacing})"
    )

    # dash_pattern
    pdf = FPDF()
    pdf.add_page()
    # verify default
    tgt_dash_pattern = dict(dash=0, gap=0, phase=0)
    assert pdf.dash_pattern == tgt_dash_pattern, (
        f"pdf.dash_pattern ({pdf.dash_pattern})"
        f" != tgt_dash_pattern ({tgt_dash_pattern})"
    )
    # change
    tgt_dash_pattern = dict(dash=2, gap=1, phase=0)
    pdf.set_dash_pattern(**tgt_dash_pattern)
    assert pdf.dash_pattern == tgt_dash_pattern, (
        f"pdf.dash_pattern ({pdf.dash_pattern})"
        f" != tgt_dash_pattern ({tgt_dash_pattern})"
    )
    # stays the same
    pdf.set_dash_pattern(**tgt_dash_pattern)
    assert pdf.dash_pattern == tgt_dash_pattern, (
        f"pdf.dash_pattern ({pdf.dash_pattern})"
        f" != tgt_dash_pattern ({tgt_dash_pattern})"
    )

    # line_width
    pdf = FPDF()
    # verify default
    tgt_line_width = 0.567 / pdf.k
    assert pdf.line_width == tgt_line_width, (
        f"pdf.line_width ({pdf.line_width})" f" != tgt_line_width ({tgt_line_width})"
    )
    # change
    tgt_line_width = 0.5
    pdf.set_line_width(tgt_line_width)
    assert pdf.line_width == tgt_line_width, (
        f"pdf.line_width ({pdf.line_width})" f" != tgt_line_width ({tgt_line_width})"
    )
    # stays the same
    pdf.set_line_width(tgt_line_width)
    assert pdf.line_width == tgt_line_width, (
        f"pdf.line_width ({pdf.line_width})" f" != tgt_line_width ({tgt_line_width})"
    )

    # text_mode
    # no setter method for this one

    # font_size_pt
    pdf = FPDF()
    # verify default
    tgt_font_size_pt = 12
    tgt_font_size = tgt_font_size_pt / pdf.k
    assert pdf.font_size_pt == tgt_font_size_pt, (
        f"pdf.font_size_pt ({pdf.font_size_pt})"
        f" != tgt_font_size_pt ({tgt_font_size_pt})"
    )
    assert pdf.font_size == tgt_font_size, (
        f"pdf.font_size ({pdf.font_size})" f" != tgt_font_size ({tgt_font_size})"
    )
    # change
    tgt_font_size_pt = 14
    tgt_font_size = tgt_font_size_pt / pdf.k
    pdf.set_font_size(tgt_font_size_pt)
    assert pdf.font_size_pt == tgt_font_size_pt, (
        f"pdf.font_size_pt ({pdf.font_size_pt})"
        f" != tgt_font_size_pt ({tgt_font_size_pt})"
    )
    assert pdf.font_size == tgt_font_size, (
        f"pdf.font_size ({pdf.font_size})" f" != tgt_font_size ({tgt_font_size})"
    )
    # stays the same
    pdf.set_font_size(tgt_font_size_pt)
    assert pdf.font_size_pt == tgt_font_size_pt, (
        f"pdf.font_size_pt ({pdf.font_size_pt})"
        f" != tgt_font_size_pt ({tgt_font_size_pt})"
    )
    assert pdf.font_size == tgt_font_size, (
        f"pdf.font_size ({pdf.font_size})" f" != tgt_font_size ({tgt_font_size})"
    )

    # font
    pdf = FPDF()
    # verify default
    tgt_font = ("", "", 12)
    assert pdf.font_family == tgt_font[0], (
        f"pdf.font_family ({pdf.font_family})" f" != tgt_font[0] ({tgt_font[0]})"
    )
    assert pdf.font_style == tgt_font[1], (
        f"pdf.font_style ({pdf.font_style})" f" != tgt_font[1] ({tgt_font[1]})"
    )
    assert pdf.font_size_pt == tgt_font[2], (
        f"pdf.font_size_pt ({pdf.font_size_pt})" f" != tgt_font[2] ({tgt_font[2]})"
    )
    # change
    for tgt_font in (
        dict(family="helvetica"),
        dict(family="helvetica", style="B"),
        dict(family="helvetica", size=14),
        dict(family="helvetica", style="I", size=12),
    ):
        pdf = FPDF()
        pdf.set_font(**tgt_font)
        assert pdf.font_family == tgt_font.get("family", ""), (
            f"pdf.font_family ({pdf.font_family})"
            f" != tgt_font.get('family', '') ({tgt_font.get('family', '')})"
        )
        assert pdf.font_style == tgt_font.get("style", ""), (
            f"pdf.font_style ({pdf.font_style})"
            f" != tgt_font.get('style', '') ({tgt_font.get('style', '')})"
        )
        assert pdf.font_size_pt == tgt_font.get("size", 12), (
            f"pdf.font_size_pt ({pdf.font_size_pt})"
            f" != tgt_font.get('size', 12) ({tgt_font.get('size', 12)})"
        )
    # stays the same
    pdf.set_font(**tgt_font)
    assert pdf.font_family == tgt_font.get("family"), (
        f"pdf.font_family ({pdf.font_family})"
        f" != tgt_font.get('family', '') ({tgt_font.get('family', '')})"
    )
    assert pdf.font_style == tgt_font.get("style"), (
        f"pdf.font_style ({pdf.font_style})"
        f" != tgt_font.get('style', '') ({tgt_font.get('style', '')})"
    )
    assert pdf.font_size_pt == tgt_font.get("size"), (
        f"pdf.font_size_pt ({pdf.font_size_pt})"
        f" != tgt_font.get('size', 12) ({tgt_font.get('size', 12)})"
    )


def test_vpos_properties():
    pdf = FPDF()
    sub_scale = 0.8
    pdf.sub_scale = sub_scale
    assert (
        pdf.sub_scale == sub_scale
    ), f"pdf.sub_scale ({pdf.sub_scale}) != sub_scale ({sub_scale})"
    sup_scale = 0.81
    pdf.sup_scale = sup_scale
    assert (
        pdf.sup_scale == sup_scale
    ), f"pdf.sup_scale ({pdf.sup_scale}) != sup_scale ({sup_scale})"
    nom_scale = 0.82
    pdf.nom_scale = nom_scale
    assert (
        pdf.nom_scale == nom_scale
    ), f"pdf.nom_scale ({pdf.nom_scale}) != nom_scale ({nom_scale})"
    denom_scale = 0.83
    pdf.denom_scale = denom_scale
    assert (
        pdf.denom_scale == denom_scale
    ), f"pdf.denom_scale ({pdf.denom_scale}) != denom_scale ({denom_scale})"
    sub_lift = -0.2
    pdf.sub_lift = sub_lift
    assert (
        pdf.sub_lift == sub_lift
    ), f"pdf.sub_lift ({pdf.sub_lift}) != sub_lift ({sub_lift})"
    sup_lift = 0.5
    pdf.sup_lift = sup_lift
    assert (
        pdf.sup_lift == sup_lift
    ), f"pdf.sup_lift ({pdf.sup_lift}) != sup_lift ({sup_lift})"
    nom_lift = 0.3
    pdf.nom_lift = nom_lift
    assert (
        pdf.nom_lift == nom_lift
    ), f"pdf.nom_lift ({pdf.nom_lift}) != nom_lift ({nom_lift})"
    denom_lift = 1
    pdf.denom_lift = denom_lift
    assert (
        pdf.denom_lift == denom_lift
    ), f"pdf.denom_lift ({pdf.denom_lift}) != denom_lift ({denom_lift})"


def test_local_context_init(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    with pdf.local_context(
        font_family="Courier", font_style="B", font_size_pt=24, text_color=(255, 128, 0)
    ):
        pdf.cell(text="Local context")
    pdf.ln()
    pdf.cell(text="Back to base")
    assert_pdf_equal(pdf, HERE / "local_context_init.pdf", tmp_path)


def test_local_context_shared_props(tmp_path):
    "Test local_context() with settings that are controlled by both GraphicsStateMixin and drawing.GraphicsStyle"
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    with pdf.local_context(
        draw_color=(0, 128, 255),
        fill_color=(255, 128, 0),
        line_width=2,
        dash_pattern=dict(dash=0.5, gap=9.5, phase=3.25),
    ):
        pdf.rect(x=60, y=60, w=60, h=60, style="DF")
    pdf.rect(x=60, y=150, w=60, h=60, style="DF")
    assert_pdf_equal(pdf, HERE / "local_context_shared_props.pdf", tmp_path)


def test_local_context_inherited_shared_props(tmp_path):
    "The only thing that should differ between the 2 squares is their opacity"
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    pdf.set_draw_color(0, 128, 255)
    pdf.set_fill_color(255, 128, 0)
    pdf.set_line_width(2)
    pdf.set_dash_pattern(dash=0.5, gap=9.5, phase=3.25)
    pdf.write(text="normal")
    with pdf.local_context(
        fill_opacity=0.5, char_vpos="SUP"
    ):  # => triggers creation of a local GraphicsStyle
        pdf.rect(x=60, y=60, w=60, h=60, style="DF")
        pdf.write(text="sup")
    pdf.rect(x=60, y=150, w=60, h=60, style="DF")
    pdf.write(text="normal")
    assert_pdf_equal(pdf, HERE / "local_context_inherited_shared_props.pdf", tmp_path)


def test_invalid_local_context_init():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(ValueError):
        with pdf.local_context(stroke_width=2):
            pass


def test_local_context_font_size_and_header_footer(tmp_path):  # issue 1204
    class PDF(FPDF):
        def header(self):
            self.cell(text=f"Header {self.page_no()}")
            self.ln()

        def footer(self):
            self.set_y(-15)
            self.cell(text=f"Footer {self.page_no()}")

    pdf = PDF()
    pdf.set_font(family="Helvetica", size=12)
    pdf.add_page()
    with pdf.local_context(font_size_pt=36):  # LABEL C
        pdf.multi_cell(w=0, text="\n".join(f"Line {i + 1}" for i in range(21)))
    assert pdf.font_size_pt == 12
    assert_pdf_equal(
        pdf, HERE / "local_context_font_size_and_header_footer.pdf", tmp_path
    )



from fpdf.drawing import GradientPaint
from fpdf.enums import GradientUnits
from fpdf.svg import SVGObject


def test_linear_gradient_parsing():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad1" x1="0%" y1="0%" x2="100%" y2="0%">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad1)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    # Check gradient was parsed
    assert len(svg_obj.gradient_definitions) == 1
    assert "#grad1" in svg_obj.gradient_definitions
    assert isinstance(svg_obj.gradient_definitions["#grad1"], GradientPaint)


def test_radial_gradient_parsing():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <radialGradient id="grad2" cx="50%" cy="50%" r="50%">
          <stop offset="0%" stop-color="yellow" />
          <stop offset="100%" stop-color="green" />
        </radialGradient>
      </defs>
      <circle cx="100" cy="100" r="80" fill="url(#grad2)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    # Check gradient was parsed
    assert len(svg_obj.gradient_definitions) == 1
    assert "#grad2" in svg_obj.gradient_definitions


def test_multiple_gradients():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad1">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
        <radialGradient id="grad2">
          <stop offset="0%" stop-color="yellow" />
          <stop offset="100%" stop-color="green" />
        </radialGradient>
      </defs>
      <rect width="100" height="200" fill="url(#grad1)" />
      <rect x="100" width="100" height="200" fill="url(#grad2)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert len(svg_obj.gradient_definitions) == 2
    assert "#grad1" in svg_obj.gradient_definitions
    assert "#grad2" in svg_obj.gradient_definitions


def test_gradient_with_multiple_stops():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="rainbow">
          <stop offset="0%" stop-color="red" />
          <stop offset="20%" stop-color="orange" />
          <stop offset="40%" stop-color="yellow" />
          <stop offset="60%" stop-color="green" />
          <stop offset="80%" stop-color="blue" />
          <stop offset="100%" stop-color="purple" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#rainbow)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert len(svg_obj.gradient_definitions) == 1
    assert "#rainbow" in svg_obj.gradient_definitions


def test_gradient_with_opacity():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="fade">
          <stop offset="0%" stop-color="red" stop-opacity="1.0" />
          <stop offset="100%" stop-color="red" stop-opacity="0.0" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#fade)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#fade" in svg_obj.gradient_definitions


def test_gradient_units_objectBoundingBox():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad" gradientUnits="objectBoundingBox">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions
    assert (
        svg_obj.gradient_definitions["#grad"].units == GradientUnits.OBJECT_BOUNDING_BOX
    )


def test_gradient_units_userSpaceOnUse():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad" gradientUnits="userSpaceOnUse" x1="0" y1="0" x2="200" y2="0">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert (
        svg_obj.gradient_definitions["#grad"].units == GradientUnits.USER_SPACE_ON_USE
    )


def test_gradient_spread_methods():
    for spread in ["pad", "reflect", "repeat"]:
        svg = f"""
        <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
          <defs>
            <linearGradient id="grad" spreadMethod="{spread}">
              <stop offset="0%" stop-color="red" />
              <stop offset="50%" stop-color="blue" />
            </linearGradient>
          </defs>
          <rect width="200" height="200" fill="url(#grad)" />
        </svg>
        """
        svg_obj = SVGObject(svg)
        assert "#grad" in svg_obj.gradient_definitions


def test_gradient_on_stroke():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="180" height="180" x="10" y="10" 
            fill="none" stroke="url(#grad)" stroke-width="10" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_with_gradientTransform():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad" gradientTransform="rotate(45)">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_radial_gradient_with_focal_point():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <radialGradient id="grad" cx="50%" cy="50%" r="50%" fx="25%" fy="25%">
          <stop offset="0%" stop-color="white" />
          <stop offset="100%" stop-color="black" />
        </radialGradient>
      </defs>
      <circle cx="100" cy="100" r="80" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_stops_with_style_attribute():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad">
          <stop offset="0%" style="stop-color:red;stop-opacity:1" />
          <stop offset="100%" style="stop-color:blue;stop-opacity:0.5" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_without_id_is_skipped():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient>
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="red" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    # Should not crash, just skip the gradient
    assert len(svg_obj.gradient_definitions) == 0


def test_gradient_without_stops_is_skipped():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="empty" />
      </defs>
      <rect width="200" height="200" fill="url(#empty)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    # Gradient should be skipped due to no stops
    assert "#empty" not in svg_obj.gradient_definitions


def test_gradient_percentage_coordinates():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad" x1="0%" y1="0%" x2="100%" y2="100%">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_absolute_coordinates():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad" gradientUnits="userSpaceOnUse" 
                        x1="0" y1="0" x2="200" y2="200">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="200" height="200" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_on_path():
    svg = """
    <svg width="200" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <path d="M 10 10 L 190 10 L 190 190 L 10 190 Z" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions


def test_gradient_on_shapes():
    svg = """
    <svg width="400" height="200" xmlns="http://www.w3.org/2000/svg">
      <defs>
        <linearGradient id="grad">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </linearGradient>
      </defs>
      <rect width="80" height="80" fill="url(#grad)" />
      <circle cx="150" cy="40" r="40" fill="url(#grad)" />
      <ellipse cx="260" cy="40" rx="40" ry="30" fill="url(#grad)" />
      <polygon points="320,10 350,80 290,80" fill="url(#grad)" />
    </svg>
    """
    svg_obj = SVGObject(svg)

    assert "#grad" in svg_obj.gradient_definitions



"""
Comprehensive test suite for unicode_range feature
Tests both the parser and the emoji fallback functionality
"""

from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.util import get_parsed_unicode_range
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent

# Unit tests for get_parsed_unicode_range method


def test_single_u_plus_codepoint():
    """Test parsing single U+ format codepoint"""

    result = get_parsed_unicode_range("U+1F600")
    assert result == {0x1F600}


def test_u_plus_range():
    """Test parsing U+ format range"""

    result = get_parsed_unicode_range("U+1F600-1F603")
    assert result == {0x1F600, 0x1F601, 0x1F602, 0x1F603}


def test_comma_separated_u_plus_ranges():
    """Test parsing comma-separated U+ ranges in single string"""

    result = get_parsed_unicode_range("U+0041-0043, U+0061-0063")
    expected = {0x41, 0x42, 0x43, 0x61, 0x62, 0x63}  # A-C, a-c
    assert result == expected


def test_list_of_u_plus_strings():
    """Test parsing list of U+ format strings"""

    result = get_parsed_unicode_range(["U+1F600-1F602", "U+2600", "U+2615"])
    expected = {0x1F600, 0x1F601, 0x1F602, 0x2600, 0x2615}
    assert result == expected


def test_integer_tuple_range():
    """Test parsing tuple with integer range"""

    result = get_parsed_unicode_range([(0x1F600, 0x1F603)])
    assert result == {0x1F600, 0x1F601, 0x1F602, 0x1F603}


def test_list_of_integer_tuples():
    """Test parsing multiple tuple ranges"""

    result = get_parsed_unicode_range([(0x41, 0x43), (0x61, 0x63)])
    expected = {0x41, 0x42, 0x43, 0x61, 0x62, 0x63}
    assert result == expected


def test_list_of_single_integers():
    """Test parsing list of individual integer codepoints"""

    result = get_parsed_unicode_range([0x1F600, 0x2600, 128515])
    assert result == {0x1F600, 0x2600, 128515}


def test_mixed_formats():
    """Test parsing mixed format input"""

    result = get_parsed_unicode_range(
        ["U+1F600-1F602", (0x2600, 0x2601), 0x2615, "U+0041"]
    )
    expected = {0x1F600, 0x1F601, 0x1F602, 0x2600, 0x2601, 0x2615, 0x41}
    assert result == expected


def test_case_insensitive_u_plus():
    """Test that u+ (lowercase) works same as U+"""

    result = get_parsed_unicode_range("u+1f600-1f602")
    expected = {0x1F600, 0x1F601, 0x1F602}
    assert result == expected


def test_whitespace_handling():
    """Test that whitespace is properly handled"""

    result = get_parsed_unicode_range("  U+1F600  ,  U+2600-2602  ")
    expected = {0x1F600, 0x2600, 0x2601, 0x2602}
    assert result == expected


def test_empty_input():
    """Test empty unicode_range returns empty set"""

    with pytest.raises(ValueError, match="unicode_range cannot be empty"):
        get_parsed_unicode_range([])


def test_single_character_range():
    """Test range where start == end"""

    result = get_parsed_unicode_range([(0x41, 0x41)])
    assert result == {0x41}


def test_large_range():
    """Test handling of very large unicode ranges"""

    result = get_parsed_unicode_range("U+4E00-9FFF")
    # Full CJK Unified Ideographs block (20,000+ characters)
    assert len(result) == 0x9FFF - 0x4E00 + 1
    assert 0x4E00 in result
    assert 0x9FFF in result


def test_supplementary_plane():
    """Test ranges in supplementary Unicode planes"""

    result = get_parsed_unicode_range("U+1F600-1F650")
    assert 0x1F600 in result
    assert 0x1F650 in result
    assert len(result) == 0x1F650 - 0x1F600 + 1


# Test error handling in get_parsed_unicode_range


def test_invalid_range_start_greater_than_end():
    """Test error when start > end in range"""

    with pytest.raises(ValueError) as error:
        get_parsed_unicode_range("U+1F603-1F600")
    assert "Invalid range" in str(error.value)


def test_invalid_tuple_length():
    """Test error for tuple with wrong number of elements"""

    with pytest.raises(ValueError) as error:
        get_parsed_unicode_range([(0x1F600, 0x1F602, 0x1F603)])
    assert "exactly 2 elements" in str(error.value)


def test_negative_codepoint():
    """Test error for negative codepoint"""

    with pytest.raises(ValueError) as error:
        get_parsed_unicode_range([-1])
    assert "must be non-negative" in str(error.value)


def test_invalid_hex_format():
    """Test error for invalid hex string"""

    with pytest.raises(ValueError):
        get_parsed_unicode_range("U+GHIJ")


def test_unsupported_type():
    """Test error for unsupported input type"""

    with pytest.raises(ValueError) as error:
        get_parsed_unicode_range([3.14])
    assert "Unsupported unicode_range item type" in str(error.value)


def test_invalid_range_format():
    """Test error for malformed range string"""

    with pytest.raises(ValueError):
        get_parsed_unicode_range("U+1F600-1F602-1F604")


# Test that unicode_range correctly filters font character maps


def test_filters_basic_latin():
    """Test filtering to basic Latin characters only"""

    pdf = FPDF()
    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0041-005A",  # A-Z only
    )
    font = pdf.fonts["dejavu"]

    # Should have A-Z
    assert 0x0041 in font.cmap  # A
    assert 0x005A in font.cmap  # Z

    # Should NOT have a-z or other chars
    assert 0x0061 not in font.cmap  # a
    assert 0x007A not in font.cmap  # z


def test_filters_multiple_ranges():
    """Test filtering with multiple ranges"""

    pdf = FPDF()
    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range=[(0x0041, 0x005A), (0x0061, 0x007A)],  # A-Z and a-z
    )
    font = pdf.fonts["dejavu"]

    assert 0x0041 in font.cmap  # A
    assert 0x0061 in font.cmap  # a
    assert 0x0030 not in font.cmap  # 0 (digit)


def test_empty_range_raises_value_error():
    """Test that an empty unicode_range raises a ValueError"""

    pdf = FPDF()
    with pytest.raises(ValueError, match="unicode_range cannot be empty"):
        pdf.add_font(family="DejaVu", fname=HERE / "DejaVuSans.ttf", unicode_range=[])


def test_none_range_keeps_full_cmap():
    """Test that None unicode_range keeps full font cmap"""

    pdf = FPDF()
    pdf.add_font(family="DejaVu", fname=HERE / "DejaVuSans.ttf", unicode_range=None)
    font = pdf.fonts["dejavu"]
    # DejaVu has thousands of glyphs
    assert len(font.cmap) > 2000


def test_preserves_font_style():
    """Test that unicode_range works with font styles"""

    pdf = FPDF()
    pdf.add_font(
        family="DejaVu",
        style="B",
        fname=HERE / "DejaVuSans-Bold.ttf",
        unicode_range="U+0041-005A",
    )
    assert "dejavuB" in pdf.fonts
    font = pdf.fonts["dejavuB"]
    assert 0x0041 in font.cmap


def test_emoji_fallback_with_unicode_range(tmp_path):
    """Test that emoji falls back to color font when main font filtered"""
    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0000-007F",  # Basic Latin only, no emoji
    )

    # Fallback: Noto Emoji renders all emojis in color
    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
    )

    pdf.set_font("DejaVu", size=24)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    # Text with emoji that DejaVu has (monochrome) but we want from NotoEmoji
    pdf.cell(w=0, text="Hello 😀 🚀 ☀ ☕")

    assert_pdf_equal(
        pdf,
        HERE / "emoji_fallback_with_unicode_range.pdf",
        tmp_path,
    )


def test_emoji_without_unicode_range_shows_monochrome(tmp_path):
    """Test baseline: without unicode_range, DejaVu's monochrome emoji is used"""

    pdf = FPDF()

    # Main font: DejaVu without unicode_range restriction
    pdf.add_font(family="DejaVu", fname=HERE / "DejaVuSans.ttf")

    # Fallback: Noto Emoji (won't be used for chars DejaVu has)
    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
    )

    pdf.set_font("DejaVu", size=24)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    # Some emoji will render as monochrome from DejaVu
    pdf.cell(w=0, text="Hello 😀 🚀 ☀ ☕")

    assert_pdf_equal(
        pdf,
        HERE / "emoji_without_unicode_range.pdf",
        tmp_path,
    )


def test_restrict_emoji_font_to_specific_ranges(tmp_path):
    """Test restricting emoji font to only handle specific emoji ranges"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0020-007E",
    )

    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
        unicode_range="U+1F600-1F64F",  # Emoticons only
    )

    pdf.set_font("DejaVu", size=24)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    # 😀 is in emoticons range (will use NotoEmoji)
    # ☀ is in misc symbols (not in NotoEmoji's restricted range as well as not in DejaVu)
    pdf.cell(w=0, text="Hello 😀 World ☀")

    assert_pdf_equal(pdf, HERE / "emoji_restricted_range.pdf", tmp_path)


def test_multiple_emoji_fonts_different_ranges(tmp_path):
    """Test using multiple emoji fonts with different unicode ranges"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0020-007E, U+2600-26FF",
    )

    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
        unicode_range="U+1F600-1F64F, U+1F680-1F6FF",
    )

    pdf.set_font("DejaVu", size=24)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    # Misc Symbols (DejaVu - monochrome)
    pdf.cell(w=0, text="Symbols: ☀ ☁ ☂ ☃ ⚡ ☕ ♠ ♥ ♦")
    pdf.ln()
    # Emoticons (NotoEmoji - colorful)
    pdf.cell(w=0, text="Faces: 😀 😁 😂 😃 😄 😅 😆 😊 😍 😎")
    pdf.ln()
    # Transport (NotoEmoji - colorful)
    pdf.cell(w=0, text="Transport: 🚀 🚁 🚂 🚃 🚄 🚅 🚗 🚙")
    pdf.ln()
    # Mixed: showing contrast
    pdf.cell(w=0, text="Mixed: ☀ 😀 ☕ 😊 ☂ 🚀 ♦ 😎 ♥")

    assert_pdf_equal(pdf, HERE / "multiple_emoji_fonts.pdf", tmp_path)


def test_mixed_text_with_multiple_scripts(tmp_path):
    """Test document with Latin, symbols, and emoji"""

    pdf = FPDF()

    # Main font: Basic Latin + Latin Extended
    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0020-007F, U+00A0-00FF, U+0100-017F",
    )

    # Emoji font
    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
        unicode_range="U+1F300-1F9FF, U+2600-27BF",
    )

    pdf.set_font("DejaVu", size=16)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    pdf.cell(w=0, text="Regular text")
    pdf.ln()
    pdf.cell(w=0, text="Accented: café, naïve, résumé")
    pdf.ln()
    pdf.cell(w=0, text="Emoji: 😀 🎉 ☀ 🚀 ⭐")
    pdf.ln()
    pdf.cell(w=0, text="Mixed: Hello 😀 café ☕")

    assert_pdf_equal(
        pdf,
        HERE / "mixed_text_multiple_scripts.pdf",
        tmp_path,
    )


def test_prefer_specialized_font_for_punctuation(tmp_path):
    """Test preferring a specialized font for smart quotes"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0020-2017, U+201A-10FFFF",  # Skip U+2018-2019
    )

    pdf.add_font(
        family="DejaVuQuotes",
        fname=HERE / "DejaVuSans-Bold.ttf",
        unicode_range="U+2018-201F",  # Smart quotes
    )

    pdf.set_font("DejaVu", size=24)
    pdf.set_fallback_fonts(["DejaVuQuotes"])
    pdf.add_page()
    pdf.cell(w=0, text="Regular text with 'smart' and \"fancy\" quotes")

    assert_pdf_equal(
        pdf,
        HERE / "specialized_punctuation.pdf",
        tmp_path,
    )


def test_multi_cell_with_unicode_range(tmp_path):
    """Test unicode_range works with multi_cell and text wrapping"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu", fname=HERE / "DejaVuSans.ttf", unicode_range="U+0020-007E"
    )

    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
        unicode_range="U+1F600-1F64F, U+2600-26FF",
    )

    pdf.set_font("DejaVu", size=16)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    long_text = (
        "This is a long text with emoji 😀 that should wrap "
        "across multiple lines ☀ and still render correctly 🚀 "
        "even when using unicode_range to control font fallback. "
        "More emoji: 😊 ⭐ 🎉"
    )

    pdf.multi_cell(w=0, text=long_text)

    assert_pdf_equal(
        pdf,
        HERE / "multi_cell_unicode_range.pdf",
        tmp_path,
    )


def test_markdown_with_unicode_range(tmp_path):
    """Test unicode_range works with markdown formatting"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu", fname=HERE / "DejaVuSans.ttf", unicode_range="U+0020-007E"
    )
    pdf.add_font(
        family="DejaVu",
        style="B",
        fname=HERE / "DejaVuSans-Bold.ttf",
        unicode_range="U+0020-007E",
    )
    pdf.add_font(
        family="DejaVu",
        style="I",
        fname=HERE / "DejaVuSans-Oblique.ttf",
        unicode_range="U+0020-007E",
    )

    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
    )

    pdf.set_font("DejaVu", size=16)
    pdf.set_fallback_fonts(["NotoEmoji"])
    pdf.add_page()

    text = "**Bold text** with emoji 😀 and __italic text__ with ☀ symbols"
    pdf.multi_cell(w=0, text=text, markdown=True)

    assert_pdf_equal(
        pdf,
        HERE / "markdown_unicode_range.pdf",
        tmp_path,
    )


def test_unicode_range_with_zero_width_characters(tmp_path):
    """Test unicode_range with zero-width and combining characters"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+0020-007F, U+0300-036F",  # Include combining diacritics
    )

    pdf.set_font("DejaVu", size=24)
    pdf.add_page()

    pdf.cell(w=0, text="e\u0301")

    assert_pdf_equal(
        pdf,
        HERE / "unicode_range_combining.pdf",
        tmp_path,
    )


def test_unicode_range_bmp_boundary(tmp_path):
    """Test unicode_range at BMP/supplementary plane boundary"""

    pdf = FPDF()

    pdf.add_font(
        family="DejaVu",
        fname=HERE / "DejaVuSans.ttf",
        unicode_range="U+FFF0-FFFF, U+10000-1000F",  # Cross plane boundary
    )

    pdf.set_font("DejaVu", size=24)
    pdf.add_page()

    # This mainly tests that parsing works across planes
    pdf.cell(w=0, text="Test")

    assert_pdf_equal(
        pdf,
        HERE / "unicode_range_bmp_boundary.pdf",
        tmp_path,
    )



from pathlib import Path

from fpdf import FPDF, FPDFException
from test.conftest import assert_pdf_equal, assert_same_file

import pytest

HERE = Path(__file__).resolve().parent

TABLE_DATA = (
    ("First name", "Last name", "Age", "City"),
    ("Jules", "Smith", "34", "San Juan"),
    ("Mary", "Ramos", "45", "Orlando"),
    ("Carlson", "Banks", "19", "Los Angeles"),
    ("Lucas", "Cimon", "31", "Angers"),
)


def test_multi_cell_table_unbreakable(tmp_path):  # issue 111
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    line_height = pdf.font_size * 2
    col_width = pdf.epw / 4  # distribute content evenly
    for i in range(5):  # repeat table 5 times
        with pdf.unbreakable() as doc:
            for row in TABLE_DATA:
                for datum in row:
                    doc.multi_cell(
                        col_width,
                        line_height,
                        f"{datum} ({i})",
                        border=1,
                        new_x="RIGHT",
                        new_y="TOP",
                    )
                doc.ln(line_height)
        pdf.ln(line_height * 2)
    assert_pdf_equal(pdf, HERE / "multi_cell_table_unbreakable.pdf", tmp_path)


def test_multi_cell_table_unbreakable2(tmp_path):  # issue 120 - 2nd snippet
    table = {
        "A": "test_lin_o_00000_001",
        "B": "3",
        "C": "4",
        "D": "test_lin_o_00000",
        "E": "7",
        "F": "test_lin_o_00000",
        "G": "test_lin",
        "H": "test_lin",
    }
    pdf = FPDF()
    pdf.add_page()
    pdf.set_margins(20, 20)
    pdf.set_font("Times", style="B", size=7)
    line_height = pdf.font_size * 3
    col_width = pdf.epw / 8
    for header in table:
        pdf.multi_cell(
            col_width,
            line_height,
            header,
            border=1,
            new_x="RIGHT",
            new_y="TOP",
            max_line_height=pdf.font_size,
            align="C",
        )
    pdf.ln(line_height)
    pdf.set_font(style="")
    line_height = pdf.font_size * 10
    col_width = pdf.epw / 8
    for _ in range(11):
        with pdf.unbreakable():
            for cell in table.values():
                pdf.multi_cell(
                    col_width,
                    line_height,
                    cell,
                    border=1,
                    new_x="RIGHT",
                    new_y="TOP",
                    max_line_height=pdf.font_size,
                    align="C",
                )
        pdf.ln(line_height)
    assert_pdf_equal(pdf, HERE / "multi_cell_table_unbreakable2.pdf", tmp_path)


def test_multi_cell_table_unbreakable_with_split_only(tmp_path):  # issue 359
    expected_warn = 'The parameter "split_only" is deprecated.'

    pdf = FPDF("P", "mm", "A4")
    pdf.set_auto_page_break(True, 20)
    pdf.set_font("Helvetica", size=10)
    pdf.add_page()

    data = (
        ("First name", "Last name", "Age", "City"),
        ("Jules", "Smith", "34", "San Juan"),
        ("Mary", "Ramos", "45", "Orlando"),
        ("Carlson", "Banks", "19", "Los Angeles"),
        ("Lucas", "Cimon", "31", "Angers"),
        ("Jules", "Smith", "34", "San Juan"),
        ("Mary", "Ramos", "45", "Orlando"),
        ("Carlson", "Banks", "19", "Los Angeles"),
        ("Lucas", "Cimon", "31", "Angers"),
        ("Jules", "Smith", "34", "San Juan"),
        ("Mary", "Ramos", "45", "Orlando"),
        ("Carlson", "Banks", "19", "Los Angeles"),
        ("Lucas", "Cimon", "31", "Angers"),
    )
    l_height = pdf.font_size * 1.2
    cell_width = pdf.epw / 4
    no_of_lines_list = []

    for row in data:
        max_no_of_lines_in_cell = 1
        for cell in row:
            with pytest.warns(DeprecationWarning, match=expected_warn) as record:
                result = pdf.multi_cell(
                    cell_width,
                    l_height,
                    cell,
                    border=1,
                    align="L",
                    new_x="RIGHT",
                    new_y="TOP",
                    max_line_height=l_height,
                    split_only=True,
                )

            for r in record:
                if r.category == DeprecationWarning:
                    assert_same_file(r.filename, __file__)

            no_of_lines_in_cell = len(result)
            if no_of_lines_in_cell > max_no_of_lines_in_cell:
                max_no_of_lines_in_cell = no_of_lines_in_cell
        no_of_lines_list.append(max_no_of_lines_in_cell)

    for j, row in enumerate(data):
        cell_height = no_of_lines_list[j] * l_height
        for cell in row:
            if j == 0:
                pdf.multi_cell(
                    cell_width,
                    cell_height,
                    "**" + cell + "**",
                    border=1,
                    fill=False,
                    align="L",
                    new_x="RIGHT",
                    new_y="TOP",
                    max_line_height=l_height,
                    markdown=False,
                )
            else:
                pdf.multi_cell(
                    cell_width,
                    cell_height,
                    cell,
                    border=1,
                    align="L",
                    new_x="RIGHT",
                    new_y="TOP",
                    max_line_height=l_height,
                )
        pdf.ln(cell_height)

    pdf.ln()

    with pytest.warns(DeprecationWarning, match=expected_warn) as record:
        with pdf.unbreakable() as doc:
            for _ in range(4):
                for row in data:
                    max_no_of_lines_in_cell = 1
                    for cell in row:
                        result = doc.multi_cell(
                            cell_width,
                            l_height,
                            cell,
                            border=1,
                            align="L",
                            new_x="RIGHT",
                            new_y="TOP",
                            max_line_height=l_height,
                            split_only=True,
                        )
                        no_of_lines_in_cell = len(result)
                        if no_of_lines_in_cell > max_no_of_lines_in_cell:
                            max_no_of_lines_in_cell = no_of_lines_in_cell
                    no_of_lines_list.append(max_no_of_lines_in_cell)

                for j, row in enumerate(data):
                    cell_height = no_of_lines_list[j] * l_height
                    for cell in row:
                        if j == 0:
                            doc.multi_cell(
                                cell_width,
                                cell_height,
                                "**" + cell + "**",
                                border=1,
                                fill=False,
                                align="L",
                                new_x="RIGHT",
                                new_y="TOP",
                                max_line_height=l_height,
                                markdown=False,
                            )
                        else:
                            doc.multi_cell(
                                cell_width,
                                cell_height,
                                cell,
                                border=1,
                                align="L",
                                new_x="RIGHT",
                                new_y="TOP",
                                max_line_height=l_height,
                            )
                    doc.ln(cell_height)

    for r in record:
        if r.category == DeprecationWarning:
            assert_same_file(r.filename, __file__)

    assert_pdf_equal(
        pdf, HERE / "multi_cell_table_unbreakable_with_split_only.pdf", tmp_path
    )


def test_unbreakable_with_local_context():  # discussion 557
    pdf = FPDF()
    pdf.set_font("Helvetica", size=10)
    pdf.add_page()
    pdf.set_y(270)  # Set position so that adding a cell triggers a page break
    with pytest.raises(FPDFException):
        with pdf.unbreakable() as doc:
            with doc.local_context(fill_opacity=0.3):
                doc.cell(doc.epw, 10, "Cell text content", border=1, fill=True)
    pdf.set_y(270)  # Set position so that adding a cell triggers a page break
    with pytest.raises(FPDFException):
        with pdf.unbreakable() as doc:
            with doc.local_context(text_color=(255, 0, 0)):
                doc.cell(doc.epw, 10, "Cell text content", border=1)


def test_unbreakable_with_get_y():  # discussion 557
    pdf = FPDF()
    pdf.set_font("Helvetica", size=10)
    pdf.add_page()
    pdf.set_y(270)  # Set position so that adding a cell triggers a page break
    with pytest.raises(FPDFException):
        with pdf.unbreakable() as doc:
            doc.cell(doc.epw, 10, f"doc.get_y(): {doc.get_y()}", border=1)



from pathlib import Path

import pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_write_page_break(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=24)
    doc.y = 20
    doc.write(text=LOREM_IPSUM)
    doc.write(text=LOREM_IPSUM)
    assert_pdf_equal(doc, HERE / "write_page_break.pdf", tmp_path)


def test_write_soft_hyphen(tmp_path):
    """
    The current behaviour is close to CSS word-break: break-all
    cf. https://developer.mozilla.org/en-US/docs/Web/CSS/overflow-wrap#comparing_overflow-wrap_word-break_and_hyphens
    We used to prefer a line break over a word split without regards to soft hyphens:
    https://github.com/py-pdf/fpdf2/blob/2.7.4/test/text/write_soft_hyphen.pdf
    But that caused issue with write_html(), cf. issue #847
    """
    s = "Donau\u00addamp\u00adfschiff\u00adfahrts\u00adgesellschafts\u00adkapitäns\u00admützen\u00adstreifen. "
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=24)
    doc.y = 20
    doc.write(text=s)
    doc.set_font("helvetica", size=24, style="B")
    doc.write(text=s)
    doc.set_font("helvetica", size=24, style="I")
    doc.write(text=s)
    doc.set_font("helvetica", size=24)
    doc.write(text=s)
    doc.set_font("helvetica", size=24, style="B")
    doc.write(text=s)
    doc.set_font("helvetica", size=24, style="I")
    doc.write(text=s)
    doc.set_font("helvetica", size=24)
    doc.write(text=s)
    assert_pdf_equal(doc, HERE / "write_soft_hyphen.pdf", tmp_path)


def test_write_trailing_nl(tmp_path):  # issue #455
    """Each item in lines triggers a line break at the end."""
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    lines = ["Hello\n", "Sweet\n", "World\n"]
    for line in lines:
        pdf.write(text=line)
    pdf.write(text="end_mmc")
    assert_pdf_equal(pdf, HERE / "write_trailing_nl.pdf", tmp_path)


def test_write_font_stretching(tmp_path):  # issue #478
    right_boundary = 60
    pdf = FPDF()
    pdf.add_page()
    # built-in font
    pdf.set_font("Helvetica", size=8)
    pdf.set_right_margin(pdf.w - right_boundary)
    pdf.write(text=LOREM_IPSUM[:100])
    pdf.ln()
    pdf.ln()
    pdf.set_stretching(150)
    pdf.write(text=LOREM_IPSUM[:100])
    pdf.ln()
    pdf.ln()
    # unicode font
    pdf.set_stretching(100)
    pdf.add_font(fname=FONTS_DIR / "DroidSansFallback.ttf")
    pdf.set_font("DroidSansFallback", size=8)
    pdf.write(text=LOREM_IPSUM[:100])
    pdf.ln()
    pdf.ln()
    pdf.set_stretching(150)
    pdf.write(text=LOREM_IPSUM[:100])
    # for reference, in lieu of a colored background
    pdf.line(pdf.l_margin, 10, pdf.l_margin, 100)
    pdf.line(right_boundary, 10, right_boundary, 100)
    assert_pdf_equal(pdf, HERE / "write_font_stretching.pdf", tmp_path)


def test_write_superscript(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=20)

    def write_this():
        pdf.write(text="2")
        pdf.char_vpos = "SUP"
        pdf.write(text="56")
        pdf.char_vpos = "LINE"
        pdf.write(text=" more line text")
        pdf.char_vpos = "SUB"
        pdf.write(text="(idx)")
        pdf.char_vpos = "LINE"
        pdf.write(text=" end")
        pdf.ln()
        pdf.ln()
        pdf.write(text="1234 + ")
        pdf.char_vpos = "NOM"
        pdf.write(text="5")
        pdf.char_vpos = "LINE"
        pdf.write(text="/")
        pdf.char_vpos = "DENOM"
        pdf.write(text="16")
        pdf.char_vpos = "LINE"
        pdf.write(text=" + 987 = x")
        pdf.ln()
        pdf.ln()
        pdf.ln()

    write_this()
    pdf.sub_scale = 0.5
    pdf.sup_scale = 0.5
    pdf.nom_scale = 0.5
    pdf.denom_scale = 0.5
    write_this()
    pdf.sub_lift = 0.0
    pdf.sup_lift = 0.0
    pdf.nom_lift = 0.0
    pdf.denom_lift = 0.0
    write_this()
    pdf.sub_lift = 1.0
    pdf.sup_lift = 1.0
    pdf.nom_lift = 1.0
    pdf.denom_lift = 1.0
    write_this()
    assert_pdf_equal(pdf, HERE / "write_superscript.pdf", tmp_path)


def test_write_char_wrap(tmp_path):  # issue #649
    right_boundary = 50
    pdf = FPDF()
    pdf.add_page()
    pdf.set_right_margin(pdf.w - right_boundary)
    pdf.set_font("Helvetica", size=10)
    pdf.write(text=LOREM_IPSUM[:200])
    pdf.ln()
    pdf.ln()
    pdf.write(text=LOREM_IPSUM[:200], wrapmode="CHAR")
    pdf.ln()
    pdf.ln()
    pdf.set_font("Courier", size=10)
    txt = "     " + "abcdefghijklmnopqrstuvwxyz" * 3
    pdf.write(text=txt)
    pdf.ln()
    pdf.ln()
    pdf.write(text=txt, wrapmode="CHAR")
    pdf.line(pdf.l_margin, 10, pdf.l_margin, 130)
    pdf.line(right_boundary, 10, right_boundary, 130)
    assert_pdf_equal(pdf, HERE / "write_char_wrap.pdf", tmp_path)


def test_write_overflow_no_initial_newline(tmp_path):  # issue-847
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font(family="Helvetica", size=20)
    pdf.write(7, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
    assert_pdf_equal(pdf, HERE / "write_overflow_no_initial_newline.pdf", tmp_path)


def test_write_empty():
    # Feeding an empty string to write() should not have any effect
    # on the internal state of the library.
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font(family="Helvetica", size=20)
    x = pdf.x
    y = pdf.y
    pdf.write(None, "")
    assert (
        pdf.x == x and pdf.y == y
    ), f"write('') has changed pdf.x ({pdf.x} from {x}) or pdf.x ({pdf.y} from {y})"


def test_write_deprecated_txt_arg():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    with pytest.warns(
        DeprecationWarning, match='The parameter "txt" has been renamed to "text"'
    ):
        # pylint: disable=unexpected-keyword-arg
        pdf.write(txt="Lorem ipsum Ut nostrud irure")



# Context: https://github.com/squidfunk/mkdocs-material/discussions/8252

from mkdocs.plugins import BasePlugin


class InvalidPageStructure(Exception):
    pass


class CheckTocPlugin(BasePlugin):
    def on_page_content(self, html, *, page, config, files):
        if len(page.toc) > 1:
            raise InvalidPageStructure(
                f'Page "{page.url}" has to many #top-level headings ({len(page.toc)}),'
                " Table Of Contents will be broken"
            )



from pathlib import Path

from fpdf import FPDF
from fpdf.image_parsing import get_img_info
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
IMAGE_PATH = HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png"


def test_full_width_image(tmp_path):  # issue-166
    img = get_img_info(IMAGE_PATH)
    pdf = FPDF(format=(img["w"], img["h"]))
    pdf.set_margin(0)
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=img["w"])
    assert_pdf_equal(pdf, HERE / "full_width_image.pdf", tmp_path)


def test_full_height_image(tmp_path):  # issue-166
    img = get_img_info(IMAGE_PATH)
    pdf = FPDF(format=(img["w"], img["h"]))
    pdf.set_margin(0)
    pdf.add_page()
    pdf.image(IMAGE_PATH, h=img["h"])
    assert_pdf_equal(pdf, HERE / "full_height_image.pdf", tmp_path)


def test_full_pdf_width_image(tmp_path):  # issue-528
    pdf = FPDF()
    pdf.set_margin(0)
    pdf.add_page()
    pdf.image(HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png", w=pdf.epw)
    assert_pdf_equal(pdf, HERE / "full_pdf_width_image.pdf", tmp_path)


def test_full_pdf_height_image(tmp_path):  # issue-528
    pdf = FPDF()
    pdf.set_margin(0)
    pdf.add_page()
    pdf.image(HERE / "png_images/51a4d21670dc8dfa8ffc9e54afd62f5f.png", h=pdf.eph)
    assert_pdf_equal(pdf, HERE / "full_pdf_height_image.pdf", tmp_path)


def test_image_with_explicit_dimensions(tmp_path):
    pdf = FPDF()
    pdf.set_margin(0)
    pdf.add_page()
    pdf.image(
        HERE / "png_images/6c853ed9dacd5716bc54eb59cec30889.png",
        w=pdf.epw,
        dims=(500, 500),
    )
    assert_pdf_equal(pdf, HERE / "image_with_explicit_dimensions.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_emoji_glyph(tmp_path):
    pdf = FPDF()

    pdf.add_font(fname=HERE / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans", size=64)
    pdf.add_page()

    pdf.multi_cell(0, text="".join([chr(0x1F600 + x) for x in range(68)]))

    pdf.set_font_size(32)
    pdf.text(10, 270, "".join([chr(0x1F0A0 + x) for x in range(15)]))

    assert_pdf_equal(pdf, HERE / "fonts_emoji_glyph.pdf", tmp_path)


def test_nb_replace(tmp_path):
    pdf = FPDF()

    pdf.add_font(fname=HERE / "DejaVuSans.ttf")
    pdf.add_page()

    pdf.set_font("DejaVuSans", size=64)
    pdf.cell(text="{nb}")

    pdf.set_font("helvetica")
    pdf.cell(text="{nb}")

    assert_pdf_equal(pdf, HERE / "fonts_remap_nb.pdf", tmp_path)


def test_two_mappings(tmp_path):
    pdf = FPDF()

    pdf.add_font(fname=HERE / "DejaVuSans.ttf")
    pdf.add_font(fname=HERE / "DroidSansFallback.ttf")
    pdf.add_page()

    pdf.set_font("DejaVuSans", size=64)
    pdf.cell(text="ABCDEF")

    pdf.set_font("DroidSansFallback")
    pdf.cell(text="DEFGHI")

    assert_pdf_equal(pdf, HERE / "fonts_two_mappings.pdf", tmp_path)


def test_thai_text(tmp_path):
    pdf = FPDF()
    pdf.add_font(fname=HERE / "Waree.ttf")
    pdf.set_font("Waree")
    pdf.add_page()
    pdf.write(text="สวัสดีชาวโลก ทดสอบฟอนต์, Hello world font test.")
    assert_pdf_equal(pdf, HERE / "thai_text.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF, FontFace
from fpdf.enums import TableSpan

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
IMG_DIR = HERE.parent / "image"


def test_table_with_rowspan(tmp_path):
    # Verify that tables with overlapping rowspans are calculated correctly
    pdf = FPDF()
    pdf.set_font("Times", size=24)

    # Test direct cell interface
    pdf.add_page()
    pdf.write(text="Defined with attributes\n\n")
    with pdf.table(text_align="CENTER", first_row_as_headings=False) as table:
        row = table.row()
        row.cell("A1", rowspan=3)
        row.cell("B1")
        row.cell("C1")
        row.cell("D1", rowspan=3)
        row = table.row()
        row.cell("B2", rowspan=4)
        row.cell("C2", rowspan=2)
        row = table.row()  # empty row fully spanned by previous rows
        row = table.row()
        row.cell("A3")
        row.cell("C3", rowspan=2)
        row.cell("D3", rowspan=2)
        row = table.row()
        row.cell("A4")

    # Test placeholder interface
    TABLE_DATA = [
        ["A1", "B1", "C1", "D1"],
        [TableSpan.ROW, "B2", "C2", TableSpan.ROW],
        [TableSpan.ROW, TableSpan.ROW, TableSpan.ROW, TableSpan.ROW],
        ["A3", TableSpan.ROW, "C3", "D3"],
        ["A4", TableSpan.ROW, TableSpan.ROW, TableSpan.ROW],
    ]
    pdf.add_page()
    pdf.write(text="Defined with items\n\n")
    with pdf.table(TABLE_DATA, text_align="CENTER", first_row_as_headings=False):
        pass

    # Test HTML interface
    pdf.add_page()
    pdf.write(text="Defined in HTML\n\n")
    pdf.write_html(
        """
        <table border=1>
        <tr align=center><td rowspan=3>A1</td><td>B1</td><td>C1</td><td rowspan=3>D1</td></tr>
        <tr align=center><td rowspan=4>B2</td><td rowspan=2>C2</td></tr>
        <tr></tr>
        <tr align=center><td>A3</td><td rowspan=2>C3</td><td rowspan=2>D3</td></tr>
        <tr align=center><td>A4</td></tr>
        </table>
    """,
        table_line_separators=True,
    )

    assert_pdf_equal(pdf, HERE / "table_with_rowspan.pdf", tmp_path)


def test_table_with_rowspan_and_colspan(tmp_path):
    # Verify that tables with both rowspans and colspans are calculated correctly
    #
    pdf = FPDF()
    pdf.set_font("Helvetica")
    options = dict(text_align="CENTER", gutter_width=5, gutter_height=5)
    red_fill = FontFace(emphasis="BIU", color=255, fill_color=(255, 0, 0))
    blue_fill = FontFace(emphasis="BOLD", color=255, fill_color=(0, 0, 255))

    # Test direct cell interface
    pdf.add_page()
    pdf.write(text="Defined with attributes\n\n")
    with pdf.table(**options) as table:
        row = table.row(["A", "B", "C", "D"])
        row = table.row()
        row.cell("A1", colspan=2, rowspan=3, style=red_fill)
        row.cell("C1", colspan=2, style=blue_fill)
        row = table.row()
        row.cell("C2", colspan=2, rowspan=2, style=red_fill)
        row = table.row()  # all columns of this row are spanned by previous rows
        row = table.row()
        row.cell("A4", colspan=4, style=blue_fill)
        row = table.row()
        row.cell("A5", colspan=2, style=blue_fill)
        row.cell("C5")
        row.cell("D5")
        row = table.row()
        row.cell("A6")
        row.cell("B6", colspan=2, rowspan=2, style=red_fill)
        row.cell("D6", rowspan=2, style=blue_fill)
        row = table.row()
        row.cell("A7")

    # Test placeholder interface
    TABLE_DATA = [
        ["A", "B", "C", "D"],
        ["A1", TableSpan.COL, "C1", TableSpan.COL],
        [TableSpan.ROW, TableSpan.ROW, "C2", TableSpan.COL],
        [TableSpan.ROW, TableSpan.ROW, TableSpan.ROW, TableSpan.ROW],
        ["A4", TableSpan.COL, TableSpan.COL, TableSpan.COL],
        ["A5", TableSpan.COL, "C5", "D5"],
        ["A6", "B6", TableSpan.COL, "D6"],
        ["A7", TableSpan.ROW, TableSpan.ROW, TableSpan.ROW],
    ]
    pdf.add_page()
    pdf.write(text="Defined with items\n\n")
    with pdf.table(TABLE_DATA, **options):
        pass

    # Test HTML interface
    # Not all options are available from HTML, but it should be close enough to verify
    HTML_DATA = """
        <table border=1 cellspacing=5 cellpadding=3>
        <tr align=center><th>A</th><th>B</th><th>C</th><th>D</th></tr>
        <tr align=center><td rowspan=3 colspan=2>A1</td><td colspan=2>C1</td></tr>
        <tr align=center><td rowspan=2 colspan=2>C2</td></tr>
        <tr></tr>
        <tr align=center><td colspan=4>A4</td></tr>
        <tr align=center><td colspan=2>A5</td><td>C5</td><td>D5</td></tr>
        <tr align=center><td>A6</td><td colspan=2 rowspan=2>B6</td><td rowspan=2>D6</td></tr>
        <tr align=center><td>A7</td></tr>
        </table>
    """
    pdf.add_page()
    pdf.write(text="Defined in HTML\n\n")
    pdf.write_html(HTML_DATA, table_line_separators=True)

    assert_pdf_equal(pdf, HERE / "table_with_rowspan_and_colspan.pdf", tmp_path)


def test_table_with_rowspan_and_pgbreak(tmp_path):
    # Verify that the rowspans interact correctly with pagebreaks
    pdf = FPDF()
    pdf.set_font("Helvetica")
    line_opts = dict(line_width=1, draw_color=(0, 255, 0))
    table_opts = dict(
        text_align="CENTER",
        headings_style=FontFace(emphasis="BOLD", fill_color=200),
        num_heading_rows=3,
    )

    # Interpreting span definitions from a string is application-dependent
    # Separation of placeholder strings from valid strings must be done in advance
    # Here is one demonstration with a markdown-style table extension
    TABLE_STRING = """
Location|<|Division A|<|<|<|Division B|<
^|^|Partition 1|<|Partition 2|<|^|^
X|Y|A|B|A|B|A|B
Internal|East|1-1|1-2|1-3|1-4|1-5|1-6
^|Central|^|2-2|^|2-4|^|2-6
^|West|3-1|^|3-3|^|3-5|^
External|East|4-1|4-2|4-3|4-4|4-5|<
^|Central|5-1|5-2|5-3|<|5-5|5-6
^|West|6-1|<|6-3|6-4|6-5|6-6
    """
    SPAN_SUB = {"<": TableSpan.COL, "^": TableSpan.ROW}
    TABLE_DATA = [
        [SPAN_SUB.get(datum, datum) for datum in row.split("|")]
        for row in TABLE_STRING.strip().splitlines()
    ]

    pdf.add_page()

    # simple table
    # with pdf.table(TABLE_DATA, **table_opts):
    with pdf.table(TABLE_DATA, **table_opts):
        pass

    # test breaking within a rowspan of the data block
    # -- verify break occurs before the offending rowspan
    # -- verify header reproduction
    pdf.set_y(pdf.h - 85)
    y0 = pdf.h - pdf.b_margin
    with pdf.local_context(**line_opts):
        pdf.line(0, y0, pdf.w, y0)
        pdf.line(0, pdf.y, pdf.w, pdf.y)
    with pdf.table(TABLE_DATA, **table_opts):
        pass

    # test breaking within the header
    # allow room for two rows of the table, but not three
    pdf.set_y(y0 - 20)
    with pdf.local_context(**line_opts):
        pdf.line(0, y0, pdf.w, y0)
        pdf.line(0, pdf.y, pdf.w, pdf.y)
    with pdf.table(TABLE_DATA, **table_opts):
        pass

    # test breaking inside a rowspan in the header
    # allow room for one row of the table, but not two
    pdf.set_y(y0 - 15)
    with pdf.local_context(**line_opts):
        pdf.line(0, y0, pdf.w, y0)
        pdf.line(0, pdf.y, pdf.w, pdf.y)
    with pdf.table(TABLE_DATA, **table_opts):
        pass

    # test breaking when there's nowhere good to do it
    # the continually overlapping rowspans mean the whole table must be kept together
    pdf.set_y(pdf.h * 2 / 3)
    with pdf.local_context(**line_opts):
        pdf.line(0, y0, pdf.w, y0)
        pdf.line(0, pdf.y, pdf.w, pdf.y)
    with pdf.table(text_align="CENTER") as table:
        table.row(["H1", "H2", "H3", "H4"])
        for i in range(15):
            row = table.row()
            for j, c in enumerate("ABCD"):
                ij = (5 + i - j) % 5
                rowspan = 3 if ij == 0 else 1
                if i <= j or ij == 0 or ij > 2:
                    row.cell(f"{c}{i}", rowspan=rowspan)
        table.row(["A15", "B15", "C15"])

    assert_pdf_equal(pdf, HERE / "table_with_rowspan_and_pgbreak.pdf", tmp_path)


def test_table_with_rowspan_images(tmp_path):
    # Verify that the rowspans interact correctly with pagebreaks
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()

    with pdf.table(text_align="CENTER", col_widths=(8, 10, 6)) as table:
        table.row(["Image", "Text", "Image"])

        row = table.row()
        row.cell(
            img=IMG_DIR / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png", rowspan=3
        )
        row.cell("One line of text")
        row.cell(img=IMG_DIR / "image_types/insert_images_insert_jpg.jpg", rowspan=2)

        table.row().cell("Two lines\nof text")

        row = table.row()
        # row.cell("Three\nlines\nof text")
        row.cell("One line of text")
        row.cell(img=IMG_DIR / "image_types/insert_images_insert_png.png", rowspan=2)

        row = table.row()
        row.cell(img=IMG_DIR / "image_types/circle.gif")
        row.cell("Two lines\nof text")

        row = table.row()
        row.cell(img=IMG_DIR / "image_types/insert_images_insert_jpg.jpg", rowspan=2)
        row.cell("One line of text")
        row.cell(img=IMG_DIR / "image_types/pythonknight.png", rowspan=2)

        row = table.row()
        row.cell("Four\nlines\nof\ntext")

    assert_pdf_equal(pdf, HERE / "table_with_rowspan_images.pdf", tmp_path)



from io import BytesIO
from pathlib import Path

import pytest
from defusedxml.common import EntitiesForbidden

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
SVG_SRCDIR = HERE.parent / "svg/svg_sources"


def test_svg_image(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox and width=100% and height=100%:
    pdf.image(SVG_SRCDIR / "SVG_logo.svg")
    assert_pdf_equal(pdf, HERE / "svg_image.pdf", tmp_path)


def test_svg_image_fixed_dimensions(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox and width=100 and height=100:
    img = pdf.image(SVG_SRCDIR / "SVG_logo_fixed_dimensions.svg")
    assert img["rendered_width"] == 100
    assert img["rendered_height"] == 100
    assert_pdf_equal(pdf, HERE / "svg_image_fixed_dimensions.pdf", tmp_path)


def test_svg_image_no_dimensions(tmp_path):
    pdf = fpdf.FPDF(format=(350, 350))
    pdf.add_page()
    # This image has a 300x300 viewbox but no width/height:
    pdf.image(SVG_SRCDIR / "SVG_logo_no_dimensions.svg")
    assert_pdf_equal(pdf, HERE / "svg_image_no_dimensions.pdf", tmp_path)


def test_svg_image_no_viewbox(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has no viewbox and width=100 and height=200:
    img = pdf.image(SVG_SRCDIR / "simple_rect_no_viewbox.svg")
    assert img["rendered_width"] == 100
    assert img["rendered_height"] == 200
    assert_pdf_equal(pdf, HERE / "svg_image_no_viewbox.pdf", tmp_path)


def test_svg_image_with_custom_width(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox and width=100% and height=100%:
    img = pdf.image(SVG_SRCDIR / "SVG_logo.svg", w=60)
    assert img["rendered_width"] == 60
    assert_pdf_equal(pdf, HERE / "svg_image_with_custom_width.pdf", tmp_path)


def test_svg_image_with_custom_width_and_no_dimensions(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox but no width/height:
    img = pdf.image(SVG_SRCDIR / "SVG_logo_no_dimensions.svg", w=60)
    assert img["rendered_width"] == 60
    assert_pdf_equal(
        pdf, HERE / "svg_image_with_custom_width_and_no_dimensions.pdf", tmp_path
    )


def test_svg_image_with_custom_width_and_no_viewbox(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has no viewbox and width=100 and height=200:
    img = pdf.image(SVG_SRCDIR / "simple_rect_no_viewbox.svg", w=60)
    assert img["rendered_width"] == 60
    assert_pdf_equal(
        pdf, HERE / "svg_image_with_custom_width_and_no_viewbox.pdf", tmp_path
    )


def test_svg_image_with_no_dimensions_and_custom_width(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox but no width/height:
    img = pdf.image(SVG_SRCDIR / "SVG_logo_no_dimensions.svg", w=60)
    assert img["rendered_width"] == 60
    assert_pdf_equal(
        pdf, HERE / "svg_image_with_no_dimensions_and_custom_width.pdf", tmp_path
    )


def test_svg_image_with_custom_size(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has a 300x300 viewbox but no width/height:
    pdf.image(SVG_SRCDIR / "SVG_logo_no_dimensions.svg", x=50, y=50, w=30, h=60)
    pdf.rect(x=50, y=50, w=30, h=60)  # Displays the bounding box
    assert_pdf_equal(pdf, HERE / "svg_image_with_custom_size.pdf", tmp_path)


def test_svg_image_with_custom_size_and_no_viewbox(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    # This image has no viewbox and width=100 and height=200:
    img = pdf.image(SVG_SRCDIR / "simple_rect_no_viewbox.svg", x=50, y=50, w=30, h=60)
    assert img["rendered_width"] == 30
    assert img["rendered_height"] == 60
    pdf.rect(x=50, y=50, w=30, h=60)  # Displaying the bounding box
    assert_pdf_equal(
        pdf, HERE / "svg_image_with_custom_size_and_no_viewbox.pdf", tmp_path
    )


def test_svg_image_no_viewbox_nor_width_and_height():
    pdf = fpdf.FPDF()
    pdf.add_page()
    fname = "simple_rect_no_viewbox_nor_width_and_height.svg"
    with pytest.raises(ValueError) as error:
        pdf.image(SVG_SRCDIR / fname)
    assert fname in str(error.value)
    with pytest.raises(ValueError) as error:
        pdf.image(SVG_SRCDIR / fname, w=60)


def test_svg_image_with_unsupported_color():
    pdf = fpdf.FPDF()
    pdf.add_page()
    fname = "unsupported-color.svg"
    with pytest.raises(ValueError) as error:
        pdf.image(SVG_SRCDIR / fname)
    assert fname in str(error.value)


def test_svg_image_style_inherited_from_fpdf(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_draw_color(255, 128, 0)
    pdf.set_fill_color(0, 128, 255)
    pdf.image(
        BytesIO(
            b'<svg width="180" height="180" xmlns="http://www.w3.org/2000/svg">'
            b'  <rect x="60" y="60" width="60" height="60" stroke-width="2"/>'
            b"</svg>"
        )
    )
    assert_pdf_equal(pdf, HERE / "svg_image_style_inherited_from_fpdf.pdf", tmp_path)


def test_svg_image_from_bytesio(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(
        BytesIO(
            b'<svg width="180" height="180" xmlns="http://www.w3.org/2000/svg">'
            b'  <rect x="60" y="60" width="60" height="60"/>'
            b"</svg>"
        )
    )
    assert_pdf_equal(pdf, HERE / "svg_image_from_bytesio.pdf", tmp_path)


def test_svg_image_from_bytes(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(
        b'<svg width="180" height="180" xmlns="http://www.w3.org/2000/svg">'
        b'  <rect x="60" y="60" width="60" height="60"/>'
        b"</svg>"
    )
    assert_pdf_equal(pdf, HERE / "svg_image_from_bytesio.pdf", tmp_path)


def test_svg_image_billion_laughs():
    "cf. https://pypi.org/project/defusedxml/#attack-vectors"
    pdf = fpdf.FPDF()
    pdf.add_page()
    with pytest.raises(EntitiesForbidden):
        pdf.image(
            BytesIO(
                b'<?xml version="1.0"?>'
                b"<!DOCTYPE lolz ["
                b'  <!ENTITY lol "lol">'
                b"  <!ELEMENT lolz (#PCDATA)>"
                b'  <!ENTITY lol1 "&lol;&lol;&lol;&lol;&lol;&lol;&lol;&lol;&lol;&lol;">'
                b'  <!ENTITY lol2 "&lol1;&lol1;&lol1;&lol1;&lol1;&lol1;&lol1;&lol1;&lol1;&lol1;">'
                b'  <!ENTITY lol3 "&lol2;&lol2;&lol2;&lol2;&lol2;&lol2;&lol2;&lol2;&lol2;&lol2;">'
                b'  <!ENTITY lol4 "&lol3;&lol3;&lol3;&lol3;&lol3;&lol3;&lol3;&lol3;&lol3;&lol3;">'
                b'  <!ENTITY lol5 "&lol4;&lol4;&lol4;&lol4;&lol4;&lol4;&lol4;&lol4;&lol4;&lol4;">'
                b'  <!ENTITY lol6 "&lol5;&lol5;&lol5;&lol5;&lol5;&lol5;&lol5;&lol5;&lol5;&lol5;">'
                b'  <!ENTITY lol7 "&lol6;&lol6;&lol6;&lol6;&lol6;&lol6;&lol6;&lol6;&lol6;&lol6;">'
                b'  <!ENTITY lol8 "&lol7;&lol7;&lol7;&lol7;&lol7;&lol7;&lol7;&lol7;&lol7;&lol7;">'
                b'  <!ENTITY lol9 "&lol8;&lol8;&lol8;&lol8;&lol8;&lol8;&lol8;&lol8;&lol8;&lol8;">'
                b"]>"
                b"<lolz>&lol9;</lolz>"
            )
        )


def test_svg_image_fit_rect(tmp_path):
    """
    Scale the image to fill the rectangle, keeping its aspect ratio,
    and ensure it does overflow the rectangle width or height in the process.
    """
    pdf = fpdf.FPDF()
    pdf.add_page()
    test_file = SVG_SRCDIR / "SVG_logo.svg"

    rect1 = 30, 30, 60, 100
    pdf.rect(*rect1)
    pdf.image(test_file, *rect1, keep_aspect_ratio=True)

    rect2 = 100, 30, 100, 60
    pdf.rect(*rect2)
    pdf.image(test_file, *rect2, keep_aspect_ratio=True)

    assert_pdf_equal(pdf, HERE / "svg_image_fit_rect.pdf", tmp_path)


IMG_DESCRIPTION = "The Mighty SVG Logo"


def test_svg_image_alt_text_title(tmp_path):
    test_file = SVG_SRCDIR / "SVG_logo.svg"
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(test_file, alt_text=IMG_DESCRIPTION, w=50, h=50)
    pdf.image(test_file, title=IMG_DESCRIPTION, w=50, h=50)
    pdf.image(test_file, alt_text=IMG_DESCRIPTION, title=IMG_DESCRIPTION, w=50, h=50)
    assert_pdf_equal(pdf, HERE / "svg_image_alt_text_title.pdf", tmp_path)


def test_svg_image_alt_text_two_pages(tmp_path):
    test_file = SVG_SRCDIR / "SVG_logo.svg"
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(test_file, alt_text=IMG_DESCRIPTION, w=50, h=50)
    pdf.add_page()
    pdf.image(test_file, alt_text=IMG_DESCRIPTION, w=50, h=50)
    assert_pdf_equal(pdf, HERE / "svg_image_alt_text_two_pages.pdf", tmp_path)



"""
Quoting the PDF spec:
> PDF’s logical _structure facilities_ provide a mechanism for incorporating
> structural information about a document’s content into a PDF file.

> The logical structure of a document is described by a hierarchy of objects called
> the _structure hierarchy_ or _structure tree_.
> At the root of the hierarchy is a dictionary object called the _structure tree root_,
> located by means of the **StructTreeRoot** entry in the document catalog.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/DocumentOutlineAndTableOfContents.html>
"""

from collections import defaultdict
from typing import TYPE_CHECKING, DefaultDict, Dict, Iterator, List, Optional, Union

from .syntax import PDFArray, PDFObject, PDFString

if TYPE_CHECKING:
    from .encryption import StandardSecurityHandler


class NumberTree(PDFObject):
    """A number tree is similar to a name tree, except that its keys are integers
    instead of strings and are sorted in ascending numerical order.

    A name tree serves a similar purpose to a dictionary—associating keys and
    values—but by different means.

    The values associated with the keys may be objects of any type. Stream objects
    are required to be specified by indirect object references. It is recommended,
    though not required, that dictionary, array, and string objects be specified by
    indirect object references, and other PDF objects (nulls, numbers, booleans,
    and names) be specified as direct objects
    """

    __slots__ = ("_id", "nums")  # RAM usage optimization

    def __init__(self) -> None:
        super().__init__()
        self.nums: DefaultDict[int, list[PDFObject]] = defaultdict(
            list
        )  # {struct_parent_id -> struct_elems}

    def serialize(
        self,
        obj_dict: Optional[Dict[str, object]] = None,
        _security_handler: Optional["StandardSecurityHandler"] = None,
    ) -> str:
        newline = "\n"
        serialized_nums = "\n".join(
            f"{struct_parent_id} [{newline.join(struct_elem.ref for struct_elem in struct_elems)}]"
            for struct_parent_id, struct_elems in self.nums.items()
        )
        return super().serialize({"/Nums": f"[{serialized_nums}]"})


class StructTreeRoot(PDFObject):
    __slots__ = ("_id", "type", "parent_tree", "k")  # RAM usage optimization

    def __init__(self) -> None:
        super().__init__()
        self.type = "/StructTreeRoot"
        # A number tree used in finding the structure elements to which content items belong:
        self.parent_tree = NumberTree()
        # The immediate child or children of the structure tree root in the structure hierarchy:
        self.k = PDFArray()


class StructElem(PDFObject):
    __slots__ = (  # RAM usage optimization
        "_id",
        "type",
        "s",
        "p",
        "k",
        "t",
        "alt",
        "pg",
        "_page_number",
    )

    def __init__(
        self,
        struct_type: str,
        parent: PDFObject,
        kids: Union[List[int], List["StructElem"]],
        page_number: Optional[int] = None,
        title: Optional[str] = None,
        alt: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.type = "/StructElem"
        # A name object identifying the nature of the structure element:
        self.s = struct_type
        self.p = parent  # The structure element that is the immediate parent of this one in the structure hierarchy
        self.k = PDFArray(kids)  # The children of this structure element
        # a text string representing it in human-readable form:
        self.t = None if title is None else PDFString(title)
        # An alternate description of the structure element in human-readable form:
        self.alt = None if alt is None else PDFString(alt)
        self.pg = None  # A page object on which some or all of the content items designated by the K entry are rendered
        self._page_number = page_number  # private so that it does not get serialized

    def page_number(self) -> Optional[int]:
        return self._page_number


class StructureTreeBuilder:
    def __init__(self) -> None:
        self.struct_tree_root = StructTreeRoot()
        self.doc_struct_elem = StructElem(
            struct_type="/Document", parent=self.struct_tree_root, kids=[]
        )
        self.struct_tree_root.k.append(self.doc_struct_elem)
        self.spid_per_page_number: Dict[int, int] = {}

    def add_marked_content(
        self,
        page_number: int,
        struct_type: str,
        mcid: Optional[int] = None,
        title: Optional[str] = None,
        alt_text: Optional[str] = None,
    ) -> tuple[StructElem, int]:
        struct_parents_id = self.spid_per_page_number.get(page_number)
        if struct_parents_id is None:
            struct_parents_id = len(self.spid_per_page_number)
            self.spid_per_page_number[page_number] = struct_parents_id
        struct_elem = StructElem(
            struct_type=struct_type,
            parent=self.doc_struct_elem,
            kids=[] if mcid is None else [mcid],
            page_number=page_number,
            title=title,
            alt=alt_text,
        )
        self.doc_struct_elem.k.append(struct_elem)
        self.struct_tree_root.parent_tree.nums[struct_parents_id].append(struct_elem)
        return struct_elem, struct_parents_id

    def next_mcid_for_page(self, page_number: int) -> int:
        return sum(
            1
            for struct_elem in self.doc_struct_elem.k
            if struct_elem.page_number() == page_number
            and struct_elem.k  # ensure it has a mcid set
        )

    def empty(self) -> bool:
        return not self.doc_struct_elem.k

    def __iter__(self) -> Iterator[PDFObject]:
        "Iterate all PDF objects in the tree, starting with the tree root"
        yield self.struct_tree_root
        yield self.doc_struct_elem
        yield self.struct_tree_root.parent_tree
        yield from self.doc_struct_elem.k



"""
Unit tests for color font palette selection feature.
Tests the palette parameter in add_font() method.
"""

from pathlib import Path
from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_palette_parameter_acceptance(tmp_path):
    """Test that the palette parameter is accepted and stored."""
    pdf = FPDF()

    # Test that palette parameter is accepted without error
    pdf.add_font(
        "Nabla", "", HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf", palette=0
    )

    # Test that we can add the font multiple times with different palettes
    pdf.add_font(
        "Nabla-P1",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=1,
    )
    pdf.add_font(
        "Nabla-P2",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=2,
    )

    # Test that we can use these fonts
    pdf.add_page()
    pdf.set_font("Nabla", size=24)
    pdf.cell(text="HELLO")

    pdf.set_font("Nabla-P1", size=24)
    pdf.cell(text="WORLD")

    pdf.set_font("Nabla-P2", size=24)
    pdf.cell(text="TEST")

    assert_pdf_equal(pdf, HERE / "palette_parameter_acceptance.pdf", tmp_path)


def test_palette_defaults(tmp_path):
    """Test that palette defaults work correctly."""
    pdf = FPDF()

    # Test default palette (None should become 0)
    pdf.add_font(
        "Nabla-Default", "", HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf"
    )

    # Test explicit palette 0
    pdf.add_font(
        "Nabla-Explicit",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=0,
    )

    # Both should work the same way
    pdf.add_page()
    pdf.set_font("Nabla-Default", size=24)
    pdf.cell(text="DEFAULT")

    pdf.set_font("Nabla-Explicit", size=24)
    pdf.cell(text="EXPLICIT")

    assert_pdf_equal(pdf, HERE / "palette_defaults.pdf", tmp_path)


def test_out_of_range_palette(tmp_path, caplog):
    """Test that out-of-range palette indices are handled gracefully with warning."""
    pdf = FPDF()

    # This should not raise an error, should fall back to palette 0 and log a warning
    pdf.add_font(
        "Nabla-OOR",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=999,
    )

    pdf.add_page()
    pdf.set_font("Nabla-OOR", size=24)
    pdf.cell(text="OUT OF RANGE")

    # Check that a warning was logged about the out-of-range palette
    assert any(
        "out of range" in record.message.lower() or "palette" in record.message.lower()
        for record in caplog.records
        if record.levelname == "WARNING"
    )

    assert_pdf_equal(pdf, HERE / "out_of_range_palette.pdf", tmp_path)


def test_multiple_palettes_same_font(tmp_path):
    """Test using multiple palettes from the same font file."""
    pdf = FPDF()

    # Add the same font with different palettes using different family names
    pdf.add_font(
        "Nabla-P0",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=0,
    )
    pdf.add_font(
        "Nabla-P1",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=1,
    )
    pdf.add_font(
        "Nabla-P2",
        "",
        HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf",
        palette=2,
    )

    pdf.add_page()

    pdf.set_font("helvetica", size=12)
    pdf.cell(text="Palette 0:", new_x="lmargin", new_y="next")
    pdf.set_font("Nabla-P0", size=20)
    pdf.cell(text="HELLO", new_x="lmargin", new_y="next")

    pdf.set_font("helvetica", size=12)
    pdf.cell(text="Palette 1:", new_x="lmargin", new_y="next")
    pdf.set_font("Nabla-P1", size=20)
    pdf.cell(text="HELLO", new_x="lmargin", new_y="next")

    pdf.set_font("helvetica", size=12)
    pdf.cell(text="Palette 2:", new_x="lmargin", new_y="next")
    pdf.set_font("Nabla-P2", size=20)
    pdf.cell(text="HELLO", new_x="lmargin", new_y="next")

    assert_pdf_equal(pdf, HERE / "multiple_palettes_same_font.pdf", tmp_path)
